#include "GA.h"

extern long TotalNumSubProbs;
extern long BoundNoSolCnt;
extern long SuccGreedyCnt;
extern long SuccLSCnt;
extern long SuccEulerCnt;
extern long FitNoSolCnt;

extern long STOCK_LEN;
extern long STOCK_WID;
extern long STOCK_AREA;

extern int subProbMethod;
extern int verbosity;
extern int algType;
extern int localSearchFlag;
extern int orderHeuristic;

extern int repairHeuristic;

extern int numItems;
extern int numTypes;
extern int tMin;
extern vector<SHAPE> shape;
extern vector<TYPE> types;

extern ofstream binTimeStream;
extern ofstream binItsStream;
extern ofstream costTimeStream;
extern ofstream costItsStream;
extern ofstream divStream;
extern ofstream unplStream;
extern ofstream resultsLog;

//*********************************************************************
inline 
int	minNumGroups(vector<vector<vector<PIECE> > > &P){
	int i, min=INT_MAX;
	for(i=0;i<P.size();i++){
		if(P[i].size() < min){
			min = P[i].size();
		}
	}
	return min;
}

//*********************************************************************
inline
void swap(int &a, int &b){
	int x=a; a=b; b=x;
}
inline
void swap(double &a, double &b){
	double x=a; a=b; b=x;
}
//*********************************************************************
inline
void relabelItems(vector<vector<PIECE> > &candSol){
	int i, j;
	vector<int> A(numTypes);
	for(i=0;i<numTypes;i++) A[i]=types[i].firstIndex;
	for(i=0;i<candSol.size();i++){
		for(j=0;j<candSol[i].size();j++){
			candSol[i][j].num = A[shape[candSol[i][j].num].shapeNum];
			A[shape[candSol[i][j].num].shapeNum]++;
		}
	}
}

//*********************************************************************
inline
double calcCost(vector<long> &stockUsage){
	int i;
	double total=0.0;
	for(i=0;i<stockUsage.size();i++){
		total += pow((double(stockUsage[i])/double(STOCK_AREA)),2);
	}
	return total/double(stockUsage.size());
}

//*********************************************************************
inline
void chooseParents(vector<int> &parent, int popSize){
	int i, r;
	//Make a random permutation of numbers 0 to popSize-1
	vector<int> A(popSize);
	for(i=0;i<popSize;i++) A[i]=i;
	for(i=A.size()-1;i>=0;i--){
		r = rand()%(i+1);
		swap(A[i],A[r]);
	}
	//The first of these are the randomly selected parents
	for(i=0;i<parent.size();i++) parent[i] = A[i];
}

//*********************************************************************
inline
void insertIntoPop(vector<vector<PIECE> > &offspring, vector<long> &offStockUsage, double &offCost, 
				vector<vector<vector<PIECE> > > &pop, vector<vector<long> > &popStockUsage, vector<double> &popCost, vector<int> &parent)
{
	//Go through the parents and ID the worst one
	int toDie, i;
	double min=DBL_MAX;
	for(i=0;i<parent.size();i++){
		if(popCost[parent[i]] < min){
			min = popCost[parent[i]];
			toDie = parent[i];
		}
	}
	//Copy osp over the parent selected toDie
	pop[toDie].swap(offspring);
	popStockUsage[toDie].swap(offStockUsage);
	popCost[toDie] = offCost;
}

//*********************************************************************
inline
void copyBestSol(vector<vector<PIECE> > &bestSoFar, vector<long> &bestStockUsage, double &bestCost, vector<vector<PIECE> > &sol, vector<long> &stockUsage, double &cost)
{	
	bestSoFar = sol;
	bestStockUsage = stockUsage;
	bestCost = cost;
}

//*********************************************************************
inline
int chooseGroup(vector<vector<int> > &SiInterX, vector<long> &offStockUsage)
{
	int i, choice=-1, numChoices=0, max=1, min = INT_MAX;
	if(repairHeuristic==1){
		//Random choice of a group/bin
		for(i=0;i<SiInterX.size();i++){
			if(SiInterX[i].size() > 0){
				if(rand()%(numChoices+1)==0){
					choice = i;	
				}
				numChoices++;
			}
		}
	} else if(repairHeuristic==2){
		//Choose bin with maximum intersection
		for(i=0;i<SiInterX.size();i++){
			if(SiInterX[i].size() >= max){
				if(SiInterX[i].size() > max) numChoices=0;
				if(rand()%(numChoices+1)==0){
					choice = i;	
					max = SiInterX[i].size();
				}
				numChoices++;
			}
		}
	} else {
		//Choose emptiest bin
		for(i=0;i<SiInterX.size();i++){
			if(SiInterX[i].size() > 0){
				if(offStockUsage[i] <= min){
					if(offStockUsage[i] < min) numChoices=0;
					if(rand()%(numChoices+1)==0){
						choice = i;
						min = offStockUsage[i];
					}
					numChoices++;
				}
			}
		}
	}
	//if choice == -1 then no choice was made -- no intersection exists. This should only occur with algorithm 4
	if(choice == -1 && algType != 4){cout<<"Error no choice was made -- no intersection exists\n"; exit(1);}
	return(choice);
}

//*********************************************************************
inline
void doMutation(vector<vector<PIECE> > &offspring, vector<long> &offStockUsage,  
				vector<PIECE> &finalStock, vector<int> &groupsToRemove, vector<vector<PIECE> > &unplaced, 
				vector<int> &missingItems, vector<long> &uStockUsage)
{
	int i, j;
	bool changedSomething = false;
	uStockUsage.clear();
	
	//Select groups to mutate and move these to unplaced
	selectGroupsToRemove(groupsToRemove,offspring.size());
	unplaced.clear();
	unplaced.resize(groupsToRemove.size());
	for(i=0; i<groupsToRemove.size(); i++){
		unplaced[i].swap(offspring[groupsToRemove[i]]);
		offStockUsage[groupsToRemove[i]] = 0;
	}
	removeAnyEmptyGroups(offspring,offStockUsage);

	//Now do the mutation
	if(localSearchFlag > 0){
		//Do mutation including local search
		do 
			changedSomething = runLS(offspring,unplaced,offStockUsage,finalStock);
		while(changedSomething == true);
		//Calculate stock usages of unplaced
		for(i=0; i<unplaced.size(); i++){
			uStockUsage.push_back(calcStockUsage(unplaced[i]));
		}
		//Now check whether groups in unplaced (if there's more than one) can be combined
		if(unplaced.size() > 1) {
			rebuild(unplaced,uStockUsage,finalStock);
		}
		//And rebuild into a full solution by just concatenating offspring and unplaced
		offspring.insert(offspring.end(), unplaced.begin(), unplaced.end());
		offStockUsage.insert(offStockUsage.end(), uStockUsage.begin(), uStockUsage.end());
	} else {
		//Do mutation by just rebuilding the solution
		missingItems.clear();
		for(i=0;i<unplaced.size();i++)
			for(j=0;j<unplaced[i].size();j++) 
				missingItems.push_back(unplaced[i][j].num);
		greedyReinsert(offspring,offStockUsage,missingItems,false,finalStock);
	}
}

//*********************************************************************
inline 
void copyBin(int x, int currentParent, vector<vector<vector<PIECE> > > &P, vector<vector<long> > &PUsage, vector<vector<vector<int> > > &PByType, 
			vector<vector<PIECE> > &offspring, vector<long> &offStockUsage, vector<int> &X, vector<int> &XList, vector<vector<int> > &SiInterX, 
			vector<vector<int> > &SiMinusSiInterX, int numParents, vector<int> &origX, vector<int> &missingItems)
{
	int i, j, k, p, origXCard, XCard;

	//First copy the selected bin P[currentParent][x] to the offspring
	offspring.push_back(P[currentParent][x]);
	offStockUsage.push_back(PUsage[currentParent][x]);
	
	//Next copy the item types from the selected bin P[currentParent][x] into the multiset origX (this uses the alternative multiset representation)
	origX.clear();
	origX.resize(numTypes,0);
	for(i=0;i<PByType[currentParent][x].size();i++) origX[PByType[currentParent][x][i]]++;
	origXCard = PByType[currentParent][x].size();
	
	//And remove the contents of bin x from P[currentParent]
	P[currentParent][x].swap(P[currentParent].back());
	P[currentParent].pop_back();
	swap(PUsage[currentParent][x],PUsage[currentParent].back());
	PUsage[currentParent].pop_back();
	PByType[currentParent][x].swap(PByType[currentParent].back());
	PByType[currentParent].pop_back();

	//Now deal with the other parents
	for(p=0;p<numParents;p++){
		if(p!=currentParent){
			//Determine a set covering of X using PByType[p] as the set of sets.
			XCard = origXCard;
			X = origX;
			while(XCard > 0){
		
				//First, construct X as an ordered list (X holds all item types to be deleted from the other parent)
				XList.clear();
				for(i=0;i<X.size();i++) for(j=0;j<X[i];j++) XList.push_back(i);

				//Prepare some containers for holding intersections
				SiInterX.clear();
				SiMinusSiInterX.clear();
				SiInterX.resize(PByType[p].size());
				SiMinusSiInterX.resize(PByType[p].size());
		
				//and construct them
				for(i=0;i<PByType[p].size();i++){
					//Calculate the intersection of bin i and X
					set_intersection(XList.begin(), XList.end(), PByType[p][i].begin(), PByType[p][i].end(), back_inserter(SiInterX[i]));
					//and calculate the difference between bin i and (the intersection of this bin and X)
					set_difference(PByType[p][i].begin(), PByType[p][i].end(), SiInterX[i].begin(), SiInterX[i].end(), back_inserter(SiMinusSiInterX[i]));
				}
		
				//Now choose a bin i in the other parent that has at least one common item type with X
				i = chooseGroup(SiInterX,PUsage[p]);

				if(algType == 3){
					//Remove appropriate items (duplicate items) from bin i of parent p and delete all appropriate items from X
					PByType[p][i].swap(SiMinusSiInterX[i]);
					for(j=0;j<SiInterX[i].size();j++){ 
						X[SiInterX[i][j]]--;
						//Because the items in the stock are in a defined order, we need to search manually for each one, and delete it
						for(k=0;k<P[p][i].size();k++){
							if(SiInterX[i][j] == shape[P[p][i][k].num].shapeNum) break;
						}
						PUsage[p][i]-=shape[P[p][i][k].num].area;
						P[p][i].erase(P[p][i].begin() + k);		
					}
					XCard-=SiInterX[i].size();
					if(PByType[p][i].empty()){
						//If bin i in parent p is now empty, we delete it
						P[p][i].swap(P[p].back());
						P[p].pop_back();
						swap(PUsage[p][i],PUsage[p].back());
						PUsage[p].pop_back();
						PByType[p][i].swap(PByType[p].back());
						PByType[p].pop_back();
					}
				} else {
					//Remove ALL items from bin i of parent p by simpy deleting the bin
					if(i==-1){
						//If we are here, there were no duplicate items, so we skip to the next iteration
						break;
					}
					P[p][i].swap(P[p].back());
					P[p].pop_back();
					swap(PUsage[p][i],PUsage[p].back());
					PUsage[p].pop_back();
					PByType[p][i].swap(PByType[p].back());
					PByType[p].pop_back();
					//Delete the appropriate items from X
					for(j=0;j<SiInterX[i].size();j++){
						X[SiInterX[i][j]]--;
					}
					XCard-=SiInterX[i].size();
				}
			}
		}
	}
}
inline
void chooseBiggestBin(vector<vector<long> > &PUsage, vector<int> &tabuList, int &x, int &y, int currentIteration)
{	
	//Chooses the most full bin from any parent that isn't tabu
	//x refers to the colour chosen, y the parent it is in. If all considered parents are empty, -1s are returned
	int i, p, max = 0, numMax = 0, numParents = tabuList.size();
	for(p=0;p<numParents;p++){
		if(tabuList[p] < currentIteration){
			//Parent p can be considered for colour transfer
			for(i=0; i<PUsage[p].size(); i++){
				if(PUsage[p][i] >= max){
					if(PUsage[p][i] > max) numMax = 0;
					if(rand()%(numMax+1)==0){
						max = PUsage[p][i];
						x = i;
						y = p;
					}
					numMax++;
				}
			}
		}
	}
	if(max==0){
		//All groups in all relevant parents are empty
		y = -1; 
		x = -1; 
	}
}
inline
void GPX(vector<vector<vector<PIECE> > > &pop, vector<vector<long> > &popStockUsage, vector<int> &parent, 
			vector<vector<PIECE> > &offspring, vector<long> &offStockUsage, vector<PIECE> &finalStock, 
			vector<vector<vector<PIECE> > > &P, vector<vector<long> > &PUsage, vector<int> &missingItems,
			vector<vector<PIECE> > &unplaced, vector<long> &uStockUsage,
			vector<vector<vector<int> > > &PByType, vector<vector<int> > &SiInterX, vector<vector<int> > &SiMinusSiInterX,
			vector<int> &X, vector<int> &XList, vector<int> &origX)
{
	//Use copies of parents at relevant positions in the population to form an offspring solution
	int i, j, group, x, currentParent, numParents=parent.size(), p;
	int Q = numParents/2;
	bool changedSomething;

	//Clear offspring so we can build a new solution
	offspring.clear();
	offStockUsage.clear();
	missingItems.clear();

	//Now make copies of the parents etc.
	for(p=0;p<numParents;p++){
		P[p] = pop[parent[p]];
		PUsage[p] = popStockUsage[parent[p]];
	}
	//As part of this, we copy parents' item types and sort the types in each bin (used with set operations below)
	PByType.clear();
	PByType.resize(numParents);
	for(p=0;p<numParents;p++){
		PByType[p].resize(P[p].size());
		for(i=0;i<P[p].size();i++){
			for(j=0;j<P[p][i].size();j++){
				PByType[p][i].push_back(shape[P[p][i][j].num].shapeNum);
			}
			sort(PByType[p][i].begin(), PByType[p][i].end());
		}
	}

	//Now decide the number of groups for the offspring
	int offSize = minNumGroups(P)-1;
	
	//Keeps a record for when parents can be considered for bin transfer. Initially all parents are available
	vector<int>	tabuList(numParents,-1);

	//Now build the offspring solution
	for(group=0;group<offSize;group++){
		//Choose the parent "currentParent" and bin x to copy next
		chooseBiggestBin(PUsage,tabuList,x,currentParent,group);
		if(x!=-1){
			//We only add a bin to the offspring if a non-empty bin from currentParent has been selected. Also update the tabuList
			tabuList[currentParent] = group + Q;
			copyBin(x,currentParent,P,PUsage,PByType,offspring,offStockUsage,X,XList,SiInterX,SiMinusSiInterX,numParents,origX,missingItems);
		}
	}
	
	//We now identify the missing items as we need to reinsert them into a solution.
	if(algType == 3){
		//If standard GPX is used then the missing items will still be in the parents. We use parent[0], though any of them are suitable
		for(i=0; i<P[0].size();i++){
			for(j=0;j<P[0][i].size();j++){
				missingItems.push_back(P[0][i][j].num);
			}
		}
	} else {
		//If the modified GPX is used then we need to go through the offspring itself to identify the missing items
		vector<int> present(numTypes, 0);
		for(i=0; i<offspring.size();i++){
			for(j=0;j<offspring[i].size();j++){
				present[shape[offspring[i][j].num].shapeNum]++;
			}
		}
		for(i=0;i<numTypes;i++){
			for(j=present[i];j<types[i].numOccurences;j++){
				missingItems.push_back(types[i].firstIndex);
			}
		}
	}
		
	if(localSearchFlag > 0){
		//We will reinsert using local search. Put all the missing items into a partial solution "unplaced" using greedy algorithm
		unplaced.clear();
		unplaced.push_back(vector<PIECE>());
		uStockUsage.clear();
		uStockUsage.push_back(0);
		greedyReinsert(unplaced,uStockUsage,missingItems,true,finalStock);
		uStockUsage.clear();
		//Now do local search on "offspring" and "unplaced" in the same way as mutation
		do 
			changedSomething = runLS(offspring,unplaced,offStockUsage,finalStock);
		while(changedSomething == true);
		//Calculate stock usages of unplaced
		for(i=0; i<unplaced.size(); i++){
			uStockUsage.push_back(calcStockUsage(unplaced[i]));
		}
		//Now check whether groups in unplaced (if there's more than one) can be combined
		if(unplaced.size() > 1){
			rebuild(unplaced,uStockUsage,finalStock);
		}
		//Finally rebuild into a full solution by just concatenating offspring and unplaced
		for(i=0; i<unplaced.size(); i++) {
			offStockUsage.push_back(uStockUsage[i]);
			offspring.push_back(unplaced[i]);
		}
	} else {
		//No local search. We just reinser using FFD
		greedyReinsert(offspring,offStockUsage,missingItems,true,finalStock);
	}
	//The offspring now contains all items. However, some item labellings may be wrong in that 
	//some item labels might appear zero times, or more than once. However, each type is present 
	//the correct number of times. The following step is optional (the code still works without it)
	relabelItems(offspring);
}

//*********************************************************************
inline
void GGA(vector<vector<vector<PIECE> > > &pop, vector<vector<long> > &popStockUsage, vector<int> &parent, 
			vector<vector<PIECE> > &offspring, vector<long> &offStockUsage, vector<PIECE> &finalStock, vector<int> &missingItems, vector<vector<PIECE> > &unplaced, 
			vector<long> &uStockUsage, vector<vector<int> > &SiInterX, vector<vector<int> > &SiMinusSiInterX, vector<int> &X, vector<int> &XList, vector<vector<int> > &offByType)
{
	//Use parents at positions parent[0] and parent[1] in the population to form an offspring solution via the GGA operator
	int i, j, left, right, XCard=0;
	bool changedSomething;

	//Offspring is initially a copy of population[parent[0]].
	offspring = pop[parent[0]];
	offStockUsage = popStockUsage[parent[0]];
	missingItems.clear();

	//Copy offspring's item types and sort the types in each bin (used with set operations below)
	offByType.clear();
	offByType.resize(offspring.size());
	for(i=0;i<offspring.size();i++){
		for(j=0;j<offspring[i].size();j++){
			offByType[i].push_back(shape[offspring[i][j].num].shapeNum);
		}
		sort(offByType[i].begin(), offByType[i].end());
	}

	//Now decide which groups to copy from the second parent. This is done using 2pt Xover on a random permutation of the groups
	permuteGroups(pop[parent[1]], popStockUsage[parent[1]]);
	left = rand()%pop[parent[1]].size();
	do {
		right = rand()%pop[parent[1]].size();
	} while (left==right);
	if(right<left) swap(left,right);
	
	//Copy the item types from the selected groups of pop[parent[1]] into the multiset X (this uses alternaitve multiset representation)
	X.clear();
	X.resize(numTypes,0);
	for(i=left+1;i<=right;i++){
		for(j=0;j<pop[parent[1]][i].size();j++) X[shape[pop[parent[1]][i][j].num].shapeNum]++;
		XCard+=pop[parent[1]][i].size();
	}
	
	//Now determine a set covering of X using offByType. Spare items are added to the list of missing items
	while(XCard > 0){
		
		//First, construct X as an ordered list (X holds all duplicate item types)
		XList.clear();
		for(i=0;i<X.size();i++) for(j=0;j<X[i];j++) XList.push_back(i);

		//Prepare some containers for holding intersections
		SiInterX.clear();
		SiMinusSiInterX.clear();
		SiInterX.resize(offspring.size());
		SiMinusSiInterX.resize(offspring.size());
		
		//and construct them
		for(i=0;i<offspring.size();i++){
			//Calculate the intersection of bin i and X
			set_intersection(XList.begin(), XList.end(), offByType[i].begin(), offByType[i].end(), back_inserter(SiInterX[i]));
			//and calculate the difference between bin i and (the intersection of this bin and X)
			set_difference(offByType[i].begin(), offByType[i].end(), SiInterX[i].begin(), SiInterX[i].end(), back_inserter(SiMinusSiInterX[i]));
		}
		
		//Now choose a bin i that has at least one common item type with X
		i = chooseGroup(SiInterX,offStockUsage);

		//And add appropriate item *indexes* to the list of missing items
		for(j=0;j<SiMinusSiInterX[i].size();j++){
			missingItems.push_back(types[SiMinusSiInterX[i][j]].firstIndex);
		}
		
		//And delete the appropriate items from X
		for(j=0;j<SiInterX[i].size();j++) X[SiInterX[i][j]]--;
		XCard-=SiInterX[i].size();
		
		//Finally, remove bin i from the offspring
		offspring.erase(offspring.begin() + i);
		offStockUsage.erase(offStockUsage.begin() + i);
		offByType.erase(offByType.begin() + i);
	}

	//Having removed the approporiate groups from the offspring, we now add the selected groups from the second parent
	for(i=left+1;i<=right;i++){
		offspring.push_back(pop[parent[1]][i]);
		offStockUsage.push_back(popStockUsage[parent[1]][i]);
	}

	//Decide whther to do local search or not.
	if(localSearchFlag > 0){
		//We will reinsert using local search. Put all the missing items into a partial solution "unplaced" using greedy algorithm
		unplaced.clear();
		unplaced.push_back(vector<PIECE>());
		uStockUsage.clear();
		uStockUsage.push_back(0);
		greedyReinsert(unplaced,uStockUsage,missingItems,true,finalStock);
		uStockUsage.clear();
		//Now do local search on "offspring" and "unplaced" in the same way as mutation
		do 
			changedSomething = runLS(offspring,unplaced,offStockUsage,finalStock);
		while(changedSomething == true);
		//Calculate stock usages of unplaced
		for(i=0; i<unplaced.size(); i++){
			uStockUsage.push_back(calcStockUsage(unplaced[i]));
		}
		//Now check whether groups in unplaced (if there's more than one) can be combined
		if(unplaced.size() > 1){
			rebuild(unplaced,uStockUsage,finalStock);
		}
		//Finally rebuild into a full solution by just concatenating offspring and unplaced
		for(i=0; i<unplaced.size(); i++) {
			offStockUsage.push_back(uStockUsage[i]);
			offspring.push_back(unplaced[i]);
		}
	} else {
		//No local search. We just reinsert using FFD
		greedyReinsert(offspring,offStockUsage,missingItems,true,finalStock);
	}
	//The offspring now contains all items. However, some item labellings may be wrong in that 
	//some item labels might appear zero times, or more than once. However, each type is present 
	//the correct number of times. The following step is optional (the code still works without it)
	relabelItems(offspring);
}

//*********************************************************************
void runGA(int targetGroups, bool usingTL, bool usingIL, int itLimit, int timeLimit, char *filename, bool diversityOn,
	int popSize, int numParents){

	if(verbosity>=1){
		if(algType == 2) cout<<"Starting Experiment with GGA; ";
		else cout<<"Starting Experiment with GA w/ GPX; ";
		cout<<"File = "<<filename<<"; SubProblemMethod = "<<subProbMethod<<"; LocalSearch = "<<localSearchFlag<<"; OrderHeuristic = "<<orderHeuristic<<"; PopSize = "<<popSize<<"; parentsPerRecombination = "<<numParents<<endl;
	}

	//POPULATION SIZE SET HERE
	int i=1, bestGroups, unplacedInCrossover;
	double bestCost, offCost, diversity;
	long timestamp;

	//MAKE POPULATION SOLUTION STRUCTURES--------------------------------------
	vector<vector<vector<PIECE> > > pop(popSize,vector<vector<PIECE> >()), P(numParents); 
	vector<vector<long> > popStockUsage(popSize, vector<long>()), PUsage(numParents);
	vector<vector<PIECE> > bestSoFar, offspring, unplaced;
	vector<long> offStockUsage, bestStockUsage, uStockUsage;
	vector<double> popCost(popSize);
	vector<vector<int> > SiInterX, SiMinusSiInterX, offByType;
	vector<vector<vector<int> > > PByType(numParents);
	vector<int> X, XList, origX, missingItems, parent(numParents);

	//DECLARE SOME CONTAINERS USED IN THE PACKING/FITTING PROCEDURES
	vector<int> groupsToRemove;
	vector<PIECE> finalStock;

	//START THE TIMER
	clock_t clockStart = clock();
	clock_t clockFinish = clockStart + (timeLimit*CLOCKS_PER_SEC);

	//Some screen output
	if(verbosity>=1) cout	<<"      It     Grp      Cost    Targ   BestGrp  BestCost            MS    NumSubProbs       Div   RecomUnplaced\n"
							<<"      --     ---      ----    ----   -------  --------            --    -----------       ---   -------------\n";

	//FORM THE INITIAL POPULATION, KEEPING TRACK OF THE BEST SO FAR AND WRITING TO THE OUTPUT FILE
	//The first solution is made using FFD and perhaps a LS-mutation
	makeInitSolution(pop[0],popStockUsage[0],finalStock,true);
	popCost[0] = calcCost(popStockUsage[0]);
	//The FFD solution is copied (as the best so far) and written to the output files
	copyBestSol(bestSoFar,bestStockUsage,bestCost,pop[0],popStockUsage[0],popCost[0]);
	bestGroups = bestSoFar.size();
	timestamp = long(((clock()-clockStart)/double(CLOCKS_PER_SEC))*1000);
	binItsStream<<bestGroups<<"\t0\n";
	binTimeStream<<bestGroups<<"\t"<<timestamp<<"\n";
	costItsStream<<bestCost<<"\t0\n";
	costTimeStream<<bestCost<<"\t"<<timestamp<<"\n";
	
	//If desired, also run LS on this solution, and update the output files as before
	if(localSearchFlag > 0){
		doMutation(pop[0],popStockUsage[0],finalStock,groupsToRemove,unplaced,missingItems,uStockUsage);
	}
	popCost[0] = calcCost(popStockUsage[0]);
	if(popCost[0]>bestCost){
		timestamp = long(((clock()-clockStart)/double(CLOCKS_PER_SEC))*1000);
		copyBestSol(bestSoFar,bestStockUsage,bestCost,pop[0],popStockUsage[0],popCost[0]);
		costItsStream<<bestCost<<"\t0\n";
		costTimeStream<<bestCost<<"\t"<<timestamp<<"\n";
		if(bestSoFar.size() < bestGroups){
			bestGroups = bestSoFar.size();
			binItsStream<<bestGroups<<"\t0\n";
			binTimeStream<<bestGroups<<"\t"<<timestamp<<"\n";
		}
	}

	//Do some output to the screen
	if(verbosity>=1){
		cout<<setw(8)<<"0"<<setw(8)<<pop[0].size()<<setw(10)<<popCost[0]<<setw(8)<<targetGroups<<setw(10)<<bestGroups<<setw(10)<<bestCost<<setw(14)<<timestamp<<setw(15)<<TotalNumSubProbs<<setw(10)<<"-"<<setw(16)<<"-\n";
	}
	
	if(bestGroups > targetGroups){
		//Need to produce more of the intitial population. These are made using FF random and possibly LS
		for(i=1;i<pop.size();i++){
			
			makeInitSolution(pop[i],popStockUsage[i],finalStock,false);
			if(localSearchFlag > 0){
				doMutation(pop[i],popStockUsage[i],finalStock,groupsToRemove,unplaced,missingItems,uStockUsage);
			}
			popCost[i] = calcCost(popStockUsage[i]);

			timestamp = long(((clock()-clockStart)/double(CLOCKS_PER_SEC))*1000);
			//Check to see if this is the best so far, and if we have reached the target number of bins/groups
			if(popCost[i]>bestCost){
				copyBestSol(bestSoFar,bestStockUsage,bestCost,pop[i],popStockUsage[i],popCost[i]);
				costItsStream<<bestCost<<"\t"<<i<<"\n";
				costTimeStream<<bestCost<<"\t"<<timestamp<<"\n";
				if(bestSoFar.size() < bestGroups){
					bestGroups = bestSoFar.size();
					binItsStream<<bestGroups<<"\t"<<i<<"\n";
					binTimeStream<<bestGroups<<"\t"<<timestamp<<"\n";
					if(bestGroups <= targetGroups) break;
				}
			}

			//Do some output to the screen
			if(verbosity>=1){
				cout<<setw(8)<<i<<setw(8)<<pop[i].size()<<setw(10)<<popCost[i]<<setw(8)<<targetGroups<<setw(10)<<bestGroups<<setw(10)<<bestCost<<setw(14)<<timestamp<<setw(15)<<TotalNumSubProbs;
				if(diversityOn && i==popSize-1){
					diversity = calcDiversity(pop);
					cout<<setw(10)<<diversity<<endl;
					divStream<<diversity<<"\n";
					unplStream<<"-\n";
				}
				else cout<<setw(10)<<"-"<<setw(16)<<"-\n";
			}

		}
	}
	
	//Do some output to the log file before continuing
	if(verbosity>=1) cout<<"Initial Population (produced in "<<timestamp<<"ms)\n";

	//NOW DECIDE IF WE WANT TO RUN THE GA
	bool stopRun = false;
	//Check to see if we should enter the main loop
	if(bestSoFar.size() <= targetGroups) stopRun = true;
	else if(usingTL && usingIL && clockFinish < clock() && i>=itLimit-1) stopRun = true;
	else if(!usingTL && usingIL && i>=itLimit-1) stopRun = true;
	else if(usingTL && !usingIL && clockFinish < clock()) stopRun = true;

	//******************************************************************************************
	//Main Loop of GA
	while(!stopRun){

		//Choose numParents different parents at random
		chooseParents(parent,popSize);

		//Do crossover of these parents to form a single offspring.
		if(algType == 2)
			GGA(pop,popStockUsage,parent,offspring,offStockUsage,finalStock,missingItems,unplaced,uStockUsage,SiInterX,SiMinusSiInterX,X,XList,offByType);
		else	
			GPX(pop,popStockUsage,parent,offspring,offStockUsage,finalStock,P,PUsage,missingItems,unplaced,uStockUsage,PByType,SiInterX,SiMinusSiInterX,X,XList,origX);

		//Keep track of how much repair we had to do in recombination
		unplacedInCrossover=missingItems.size();

		//Evaluate the offspring and see if its cost is the best so far
		offCost = calcCost(offStockUsage);
		if(offCost>bestCost){
			timestamp = long(((clock()-clockStart)/double(CLOCKS_PER_SEC))*1000);
			copyBestSol(bestSoFar,bestStockUsage,bestCost,offspring,offStockUsage,offCost);
			costItsStream<<bestCost<<"\t"<<i<<"\n";
			costTimeStream<<bestCost<<"\t"<<timestamp<<"\n";
			if(bestSoFar.size() < bestGroups){
				bestGroups = bestSoFar.size();
				binItsStream<<bestGroups<<"\t"<<i<<"\n";
				binTimeStream<<bestGroups<<"\t"<<timestamp<<"\n";
			}
		}

		//Now mutate the offspring
		doMutation(offspring,offStockUsage,finalStock,groupsToRemove,unplaced,missingItems,uStockUsage);
		
		//Again, evaluate the offspring and see if its cost is the best so far
		offCost = calcCost(offStockUsage);
		if(offCost>bestCost){
			timestamp = long(((clock()-clockStart)/double(CLOCKS_PER_SEC))*1000);
			copyBestSol(bestSoFar,bestStockUsage,bestCost,offspring,offStockUsage,offCost);
			costItsStream<<bestCost<<"\t"<<i<<"\n";
			costTimeStream<<bestCost<<"\t"<<timestamp<<"\n";
			if(bestSoFar.size() < bestGroups){
				bestGroups = bestSoFar.size();
				binItsStream<<bestGroups<<"\t"<<i<<"\n";
				binTimeStream<<bestGroups<<"\t"<<timestamp<<"\n";
			}
		}

		//Insert offspring into the population by replacing the weakest parent
		insertIntoPop(offspring,offStockUsage,offCost,pop,popStockUsage,popCost,parent);
		
		//Write stuff to the screen
		if(verbosity>=1){
			timestamp = long(((clock()-clockStart)/double(CLOCKS_PER_SEC))*1000);
			cout<<setw(8)<<i<<setw(8)<<offspring.size()<<setw(10)<<offCost<<setw(8)<<targetGroups<<setw(10)<<bestGroups<<setw(10)<<bestCost<<setw(14)<<timestamp<<setw(15)<<TotalNumSubProbs;
			if(diversityOn){
				diversity = calcDiversity(pop);
				cout<<setw(10)<<diversity<<endl;
				divStream<<diversity<<"\n";
				unplStream<<unplacedInCrossover<<"\n";
			} 
			else cout<<setw(10)<<"-"<<setw(15)<<unplacedInCrossover<<endl;
		}

		//Check to see if we should end
		if(bestSoFar.size() <= targetGroups) stopRun = true;
		else if(usingTL && usingIL && clockFinish < clock() && i>=itLimit-1) stopRun = true;
		else if(!usingTL && usingIL && i>=itLimit-1) stopRun = true;
		else if(usingTL && !usingIL && clockFinish < clock()) stopRun = true;

		i++;
	}
	//******************************************************************************************
	
	clockFinish = clock();
	long duration;
	duration = long(((clockFinish-clockStart)/double(CLOCKS_PER_SEC))*1000);

	//Check that the solution is valid (bug check)
	checkSolution(bestSoFar,bestStockUsage);

	//Write solution for the visualiser
	prettyPrintOutputSolution(bestSoFar,bestStockUsage);

	//Output stuff to screen if needed
	if(verbosity>=1){
		if(bestSoFar.size() == targetGroups) cout<<"Solution found that uses target number of bins"<<endl;
		cout<<"\n\nFinal Solution ("<<i<<" iterations):";
		prettyPrintSolution(bestSoFar,bestStockUsage);
		if(verbosity>=2) prettyPrintSolutionProj(bestSoFar,bestStockUsage);
		cout<<"\n\n----------------\nSummary\n----------------"<<endl;
		cout<<"TotalNumSubProbs = "<<TotalNumSubProbs<<endl;
		cout<<"BoundNoSolCnt    = "<<BoundNoSolCnt<<endl;
		cout<<"SuccGreedyCnt    = "<<SuccGreedyCnt<<endl;
		cout<<"SuccLSCnt        = "<<SuccLSCnt<<endl;
		cout<<"SuccEulerCnt     = "<<SuccEulerCnt<<endl;
		cout<<"FitNoSolCnt      = "<<FitNoSolCnt<<endl;
		cout<<"----------------"<<endl;
		cout <<"Time taken    = "<< duration <<" ms of CPU time"<<endl;
	}
	if(TotalNumSubProbs != SuccGreedyCnt + SuccEulerCnt + BoundNoSolCnt + SuccLSCnt + FitNoSolCnt)
		cout<<"Error: Some subproblems were unaccounted for\n";

	//Output Results and close output streams
	resultsLog<<bestSoFar.size()<<"\t"<<bestCost<<"\t"<<duration<<"\t"<<i<<"\t"<<TotalNumSubProbs<<"\t"<<BoundNoSolCnt<<"\t"<<SuccGreedyCnt<<"\t"<<SuccLSCnt<<"\t"<<SuccEulerCnt<<"\t"<<FitNoSolCnt<<"\n";
	binTimeStream<<bestSoFar.size()-1<<"\tX\t"<<duration<<"\n";
	binItsStream<<bestSoFar.size()-1<<"\tX\t"<<i<<"\n";
	costTimeStream<<"1.0\tX\t"<<duration<<"\n";
	costItsStream<<"1.0\tX\t"<<i<<"\n";


	if(verbosity>=1){
		if(algType == 2) cout<<"Finished Experiment with GGA; ";
		else cout<<"Finsished Experiment with GA w/ GPX; ";
		cout<<"File = "<<filename<<"; SubProblemMethod = "<<subProbMethod<<"; LocalSearch = "<<localSearchFlag<<"; OrderHeuristic = "<<orderHeuristic<<endl;
	}

}


