#include "diversity.h"

extern int numItems;
extern int numTypes;
extern int tMin;
extern vector<SHAPE> shape;
extern vector<TYPE> types;

//-------------------------------------------------------------------------------------
double solDistance(vector<vector<PIECE> > &sol1, vector<vector<PIECE> > &sol2){
	//Calculate the solution distance between Sol1 and Sol2 using Sorensen/Dice measure
	int i,j,k,tmp;
	vector<vector<int> > ms1, ms2, result;
	vector<int> pair(2);

	//First build up multisets of pairs for each solution 
	for(k=0;k<sol1.size();k++){
		if(sol1[k].size()>=2){
			//Add all pairs of item types in the same groups in sol1 to the multiset ms1
			for(i=0;i<sol1[k].size()-1;i++){
				for(j=i+1;j<sol1[k].size();j++){
					pair[0] = shape[sol1[k][i].num].shapeNum; 
					pair[1] = shape[sol1[k][j].num].shapeNum;
					if(pair[0] > pair[1]){
						//flip the values to keep the pair ordered
						tmp=pair[0]; pair[0]=pair[1]; pair[1]=tmp;
					}
					ms1.push_back(pair);
				}
			}
		}
	}

	for(k=0;k<sol2.size();k++){
		if(sol2[k].size()>=2){
			//Add all pairs of item types in the same groups in sol2 to the multiset ms2
			for(i=0;i<sol2[k].size()-1;i++){
				for(j=i+1;j<sol2[k].size();j++){
					pair[0] = shape[sol2[k][i].num].shapeNum;
					pair[1] = shape[sol2[k][j].num].shapeNum;
					if(pair[0] > pair[1]){
						//flip the values to keep the pair ordered
						tmp=pair[0]; pair[0]=pair[1]; pair[1]=tmp;
					}
					ms2.push_back(pair);
				}
			}
		}
	}
	
	//Identify the multiset, called "result", which is the intersection of multisets ms1 and ms2
	sort(ms1.begin(), ms1.end());
	sort(ms2.begin(), ms2.end());
	set_intersection(ms1.begin(), ms1.end(), ms2.begin(), ms2.end(), back_inserter(result));

	//Return the Sorensen Dice Measure as a distance (i.e. 1-similarity)
	return(1 - ((2.0*result.size()) / double(ms1.size() + ms2.size())));
}

//-------------------------------------------------------------------------------------
double calcDiversity(vector<vector<vector<PIECE> > > &pop){
	int i, j, n=(pop.size()*(pop.size()-1))/2;
	double total=0.0;
	for(i=0;i<pop.size()-1;i++){
		for(j=i+1;j<pop.size();j++){
			total += solDistance(pop[i],pop[j]);
		}
	}
	return total/double(n);
}
