#include "fns.h"

extern long STOCK_LEN;
extern long STOCK_WID;
extern long STOCK_AREA;

extern int subProbMethod;
extern int verbosity;
extern int algType;
extern int localSearchFlag;
extern int orderHeuristic;
extern int repairHeuristic;

extern int numItems;
extern int numTypes;
extern int tMin;
extern vector<SHAPE> shape;
extern vector<TYPE> types;

//-------------------------------------------------------------------------------------
long calcStockUsage(vector<PIECE> &theItems){
	int i;
	long total = 0;
	for(i=0; i<theItems.size(); i++) {
		total += shape[theItems[i].num].area;
	}
	return(total);
}

//-------------------------------------------------------------------------------------
void removeAnyEmptyGroups(vector<vector<PIECE> > &candSol, vector<long> &stockUsage){
	int i = 0;
	while(i<candSol.size()){
		if(candSol[i].empty() && stockUsage[i]<=0){
			candSol.erase(candSol.begin() + i);
			stockUsage.erase(stockUsage.begin() + i);
		}
		else i++;
	}
}

//-------------------------------------------------------------------------------------
void permuteGroups(vector<vector<PIECE> > &candSol, vector<long> &stockUsage){
	int i, ran;
	long tempD;
	for(i = candSol.size()-1; i>=0; i--){
		ran = rand()%(i+1);
		//now we swap the two groups
		candSol[i].swap(candSol[ran]);
		tempD = stockUsage[i];	stockUsage[i] = stockUsage[ran]; stockUsage[ran] = tempD;
	}
}
void permuteGroups(vector<vector<PIECE> > &candSol){
	//Overloaded version used when the bins' fullnesses are not recorded
	int i, ran;
	for(i = candSol.size()-1; i>=0; i--){
		ran = rand()%(i+1);
		candSol[i].swap(candSol[ran]);
	}
}
//-------------------------------------------------------------------------------------
//Functions for "greedyReinsert". This Takes a partial solution "candSol" and a list "A" of unplaced items, and uses the greedy algorithm to form a new complete solution
//If we use FFD, then the items are sorted by area via quicksort, else they are sorted randomly
inline
int PartitionByArea(vector<int> &A, int low, int high)
{
	int p = low, i;
	for(i = p+1; i<=high; i++){
		if(shape[A[i]].area > shape[A[p]].area){
			swap(A[i], A[p]);
			if (i != p+1) swap(A[i], A[p+1]);
			p++;
		}
	}
	return p;
}
inline
void QuickSortByArea(vector<int> &A, int low, int high)
{
	if (low >= high) return;
	int p = PartitionByArea(A,low,high);
	QuickSortByArea(A,low,p-1);
	QuickSortByArea(A,p+1,high);
}
inline
void RandPermute(vector<int> &A, int left, int right)
{
	//Randomly permutes items between position left (inclusive) and right (not inclusive)
	int i, ran;
	for(i=right-1; i>=left; i--){
		ran = rand()%(i+1);
		swap(A[i],A[ran]);
	}
}
void greedyReinsert(vector<vector<PIECE> > &candSol, vector<long> &stockUsage, vector<int> &A, bool sortDescending,
				vector<PIECE> &finalStock){
	int i, g;
	bool inserted;
	PIECE theNewPiece;

	//Decide how the items should be reordered before applying the greedy algorithm
	if(sortDescending){
		QuickSortByArea(A,0,A.size()-1);
	} else {
		if(orderHeuristic == 1){
			RandPermute(A,0,A.size());
		} else {
			QuickSortByArea(A,0,A.size()-1);
			//ID point where items are less than 50% area
			for(i=0;i<A.size();i++){
				if(shape[i].area < STOCK_AREA/2) break;
			}
			if(i>1) RandPermute(A,0,i);
			RandPermute(A,i,A.size());
		}
	}

	//Now reinsert the items using the greedy algortihm
	for(i=0; i<A.size(); i++){
		g=0;
		inserted = false;
		while(g<candSol.size() && !inserted){
			if(packingPossible(A[i],candSol,g,stockUsage,finalStock)){
				//The item is added to the stock g
				inserted = true;
			}
			else g++;
		}
		if(!inserted){
			//If we are here then we need to create a new stock
			candSol.push_back(vector<PIECE>());
			theNewPiece.num=A[i];
			theNewPiece.LtoR=true;
			candSol[candSol.size()-1].push_back(theNewPiece);
			stockUsage.push_back(shape[theNewPiece.num].area);		
		}
	}
}

//-------------------------------------------------------------------------------------
//Function for selecting which groups to remove for Mutation, LocalSearch etc.
void selectGroupsToRemove(vector<int> &groupsToRemove, int numGroups){
	int i;
	groupsToRemove.clear();

	//One random group is removed automatically
	groupsToRemove.push_back(rand()%numGroups);

	//Go through remaining groups and remove each one with a fixed probability 3/numGroups
	for(i=0;i<groupsToRemove[0];i++){
		if(rand()/double(RAND_MAX) <= 3.0/double(numGroups))
			groupsToRemove.push_back(i);
	}
	for(i=groupsToRemove[0]+1;i<numGroups;i++){
		if(rand()/double(RAND_MAX) <= 3.0/double(numGroups))
			groupsToRemove.push_back(i);
	}

	//If too many groups have been selected for removal (more than half) then we start again. We choose a number between
	//1 and numGroups/2 and then select this many groups
	if(groupsToRemove.size() > numGroups/2){
		groupsToRemove.clear();
		int len = (rand()%(numGroups/2))+1;
		vector<int> A(numGroups);
		for(i=0;i<numGroups;i++)A[i]=i;
		RandPermute(A,0,numGroups);
		for(i=0;i<len;i++)groupsToRemove.push_back(A[i]);
	}
}


//-------------------------------------------------------------------------------------
//Local search procedure for increasing the fullness of partialCol by emptying U.
bool runLS(vector<vector<PIECE> > &partialSol, vector<vector<PIECE> > &U, vector<long> &stockUsage, 
			vector<PIECE> &finalStock)
{
	bool changedSomething;
	int g, h, m, n, ka, j;
	long pairSize, uSize, sinSize;
	//Permute the partial solution to add additional randomness
	permuteGroups(partialSol,stockUsage);
	permuteGroups(U);
	//Now run the LS procedure until nothing changes (we are at a local optimum)
PAIRPAIR:
	changedSomething = false;
	for(g=0; g<partialSol.size(); g++){
		//Check whether 2 items in stock g can be replaced by two items from a stock h in U such that usage in g increases
		if( (partialSol[g].size() >= 2) && (stockUsage[g] < STOCK_AREA) ){
			//Go through each pair of items in group g
			for(j=0; j<partialSol[g].size()-1; j++){
				for(ka=j+1; ka<partialSol[g].size(); ka++){
					pairSize = shape[partialSol[g][j].num].area + shape[partialSol[g][ka].num].area;
					//Look to to see if there is a pair of items in U[h] that have a combined size larger than "pairSize"
					for(h=0; h<U.size(); h++){
						if(U[h].size() >=2){
							//Go thru each pair in group U[h]
							for(m=0; m<U[h].size()-1; m++){
								for(n=m+1; n<U[h].size(); n++){
									uSize = shape[U[h][m].num].area + shape[U[h][n].num].area;
									//We now want to decide if we want to swap these pairs of items
									if(uSize > pairSize && ((stockUsage[g]-pairSize)+uSize) <= STOCK_AREA){
										if(doPairPairSwap(partialSol,g,j,ka,U,h,m,n,pairSize,uSize,stockUsage)){										
											changedSomething = true;
											goto PAIRSIN;
										}
									}
								}
							}
						}
					}
				}
			}
		}
	}
PAIRSIN:
	for(g=0; g<partialSol.size(); g++){
		//Now check whether 2 items in stock g can be replaced by 1 items in U such that usage in g is greater than or equal
		if(partialSol[g].size() >= 2){
			for(j=0; j<partialSol[g].size()-1; j++){
				for(ka=j+1; ka<partialSol[g].size(); ka++){
					pairSize = shape[partialSol[g][j].num].area + shape[partialSol[g][ka].num].area;
					//Look to see if there is an item m in any group h in U that has a size larger than or equal to pair size
					for(h=0;h<U.size();h++){
						for(m=0; m<U[h].size(); m++){
							uSize = shape[U[h][m].num].area;
							//We now want to decide if we want to swap the pair with the single item
							if(uSize >= pairSize && ((stockUsage[g]-pairSize)+uSize) <= STOCK_AREA){
								if(doPairSinSwap(partialSol,g,j,ka,U,h,m,pairSize,uSize,stockUsage)){
									changedSomething=true;
									goto SINSIN;
								}
							}
						}
					}
				}
			}
		}
	}
SINSIN:
	for(g=0; g<partialSol.size(); g++){
		//Now check whether 1 item in stock g can be replaced by 1 item in U
		if(stockUsage[g] < STOCK_AREA ){
			for(j=0; j<partialSol[g].size(); j++){
				for(h=0;h<U.size();h++){
					for(m=0; m<U[h].size(); m++){
						sinSize = shape[partialSol[g][j].num].area;
						uSize = shape[U[h][m].num].area;
						if(uSize > sinSize && ((stockUsage[g]-sinSize)+uSize) <= STOCK_AREA){
							//We now want to decide if we want to swap these items
							if(doSinSinSwap(partialSol,g,j,U,h,m,sinSize,uSize,stockUsage)){
								changedSomething = true;
							}
						}
					}
				}
			}
		}
	}
	//Check if any items in U can be moved into an existing group in partialSol. If this results in an empty group, delete the group	
	for(h=0;h<U.size();h++){
		for(m=0; m<U[h].size(); m++){
			for(g=0; g<partialSol.size(); g++){
				if(stockUsage[g] + shape[U[h][m].num].area <= STOCK_AREA){
					if(packingPossible(U[h][m].num,partialSol,g,stockUsage,finalStock)){
						//Item at U[h][m] can be inserted into group g, so we do it and then chck of this makes group h emty
						U[h].erase(U[h].begin() + m);
						if(U[h].empty()) U.erase(U.begin() + h);
						changedSomething = true;
						goto PAIRPAIR;
					}
				}
			}
		}
	}
	if(changedSomething){
		goto PAIRPAIR;
	}
	else return false;
}

//-------------------------------------------------------------------------------------
//Functions for "rebuild" procedure, which takes a complete feasible solution, shuffles the groups,
//and then applies the greedy procedure. (This does not increase the number of groups, but may decrease it
inline
void greedyInsert(vector<vector<PIECE> > &oldSol, vector<long> &oldStockUsage, 
			vector<PIECE> &finalStock){
	int i;
	PIECE thePiece;
	//This procedure takes a pre-grouped candidate solution and applies the greedy
	//algorithm. The new solution is copied into a new vector and this is returned	

	//First we create a new structure to hold the newSol. This has as many groups as the 
	//old solution to start. However, it could be the case that some of these might be empty at
	//the end. If they are, then they are deleted.
	vector<vector<PIECE> > newSol(oldSol.size());
	vector<long> newStockUsage(oldSol.size(), 0);
	
	//First, copy the first group into the new solution
	newSol[0] = oldSol[0];
	newStockUsage[0] += oldStockUsage[0];

	//Now we go through the remaining groups in oldSol. We take each item in turn and see if
	//it is suitable for any existing bin. If it isn't, we shove it in the end bin (we don't need to 
	//check if this is suitable
	int oldgroup, g;
	bool inserted;
	for(oldgroup = 1; oldgroup<oldSol.size(); oldgroup++){
		for(i=0; i<oldSol[oldgroup].size(); i++){
			//We now check all of the groups g up to "oldGroup-1" in newSol to see if item "oldSol[oldgroup][i]" can be 
			//inserted here. If it can, we do; if it cannot we simply shove it into newSol[oldGroup];
			g = 0;
			inserted = false;
			while(g < oldgroup && !inserted){
				if(newSol[g].empty()){
					//The stock is currently empty so the item is added to the stock g
					thePiece.num=oldSol[oldgroup][i].num;
					thePiece.LtoR=true;
					newSol[g].push_back(thePiece);
					newStockUsage[g] = shape[thePiece.num].area;
					inserted = true;
				}
				else if(packingPossible(oldSol[oldgroup][i].num,newSol,g,newStockUsage,finalStock)){
					//The stock isn't empty so we check if there is enough room and add to stock g if possible
					inserted = true;
				}
				else {
					//Packing not possible so go on to the next group
					g++;
				}
			}
			if(!inserted){
				//If we are here then the item could not be inserted into any of the groups g below "oldgroup"
				//so we just stick it into the group "oldgroup" that it came from
				PIECE thePiece;
				thePiece.num=oldSol[oldgroup][i].num;
				thePiece.LtoR=oldSol[oldgroup][i].LtoR;
				newSol[oldgroup].push_back(thePiece);
				newStockUsage[oldgroup] += shape[thePiece.num].area;					
			}
		}
	}
	removeAnyEmptyGroups(newSol,newStockUsage);

	//Finally we delete the old version and replace it with the new
	newSol.swap(oldSol);
	oldStockUsage.swap(newStockUsage);
}
void rebuild(vector<vector<PIECE> > &candSol, vector<long> &stockUsage, 
			vector<PIECE> &finalStock)
{
	//Randomly reorder the groups and then rebuild this solution using the resultant permutation
	permuteGroups(candSol,stockUsage);
	greedyInsert(candSol,stockUsage,finalStock);
}

//-------------------------------------------------------------------------------------
//Function for making an initial solution from scratch, either by FFD heurisic, or the FFR heuristic
void makeInitSolution(vector<vector<PIECE> > &candSol, vector<long> &stockUsage,
					vector<PIECE> &finalStock, bool sortDescending)
{
	int i;
	candSol.clear();
	stockUsage.clear();
	
	//1) Make an empty vector representing the unplaced items (i.e. all of them).
	vector<int> A(numItems);
	for (i=0;i<numItems;i++)A[i]=i;
	
	//2) Open up the first bin
	candSol.push_back(vector<PIECE>());
	stockUsage.push_back(0);

	//Build the solution using first-fit descending (FFD) or first-fit random (FFR)
	greedyReinsert(candSol,stockUsage,A,sortDescending,finalStock);
}
