#include "io.h"

extern long STOCK_LEN;
extern long STOCK_WID;
extern long STOCK_AREA;

extern int subProbMethod;
extern int verbosity;
extern int algType;
extern int localSearchFlag;
extern int orderHeuristic;
extern int repairHeuristic;

extern int numItems;
extern int numTypes;
extern int tMin;
extern vector<SHAPE> shape;
extern vector<TYPE> types;

//-------------------------------------------------------------------------------------
inline
int max(int &a, int &b){
	if(a>=b)
		return a;
	else return(b);
}

//-------------------------------------------------------------------------------------
void readInputFile(ifstream &inStream)
{
	int i,j,cnt=0;
	double e,f;
	long totalArea=0;
		
	//First read in the details of all the item types
	for(i=0; i<numTypes; i++){
		inStream>>types[i].numOccurences>>types[i].len>>types[i].p1>>types[i].p2>>e>>f>>types[i].parallelogram;
		types[i].area = (STOCK_WID*types[i].len) - ((STOCK_WID*types[i].p1)/2) - ((STOCK_WID*types[i].p2)/2);
		types[i].centralLen = types[i].len - (types[i].p1 + types[i].p2);
		types[i].firstIndex=cnt;
		//Simple error check
		if(types[i].len>STOCK_LEN){cout<<"Error: there is a stock longer than the maximum STOCK_LEN. Exiting...\n"; exit(1); }
		totalArea += types[i].area * types[i].numOccurences;
		cnt+=types[i].numOccurences;
	}

	//Now use the vector of types to build up the actual problem (multiple occurences of each shape)
	cnt=0;
	for(i=0; i<numTypes; i++){
		for(j=0;j<types[i].numOccurences;j++){
			shape[cnt].len = types[i].len;
			shape[cnt].centralLen = types[i].centralLen;
			shape[cnt].area = types[i].area;
			shape[cnt].p1 = types[i].p1;
			shape[cnt].p2 = types[i].p2;
			shape[cnt].shapeNum = i;
			shape[cnt].parallelogram = types[i].parallelogram;
			cnt++;
		}
	}
	//This is a bit cheeky but here we define an extra dummy shape used as part of the Euler algorithm------
	//Note that this shape is not part of the problem, thus we only usually consider items 0 to numItems-1 inclusive
	shape[cnt].len = 0;
	shape[cnt].centralLen = 0;
	shape[cnt].area = 0;
	shape[cnt].p1 = 0;
	shape[cnt].p2 = 0;
	shape[cnt].shapeNum = -1;
	
	tMin = int(ceil(totalArea/double(STOCK_LEN*STOCK_WID)));
}

//-------------------------------------------------------------------------------------
inline
void printStock(vector<PIECE> &S){
	int i;
	cout<<"Sln = ( ";
	for(i=0; i<S.size()-1; i++){
		cout << "(" << S[i].num<< ",";
		if(S[i].LtoR)
			cout<<"-->, "<<shape[S[i].num].p1<<", "<<shape[S[i].num].centralLen<<", "<<shape[S[i].num].p2<<", ";
		else 
			cout<<"<--, "<<shape[S[i].num].p2<<", "<<shape[S[i].num].centralLen<<", "<<shape[S[i].num].p1<<", ";
		cout << shape[S[i].num].area<<"), ";
	}
	cout << "(" << S[i].num<< ",";
	if(S[i].LtoR)
		cout<<"-->, "<<shape[S[i].num].p1<<", "<<shape[S[i].num].centralLen<<", "<<shape[S[i].num].p2<<", ";
	else 
		cout<<"<--, "<<shape[S[i].num].p2<<", "<<shape[S[i].num].centralLen<<", "<<shape[S[i].num].p1<<", ";
	cout << shape[S[i].num].area<<") )";
}

//-------------------------------------------------------------------------------------
inline
int calcExpanse(vector<vector<PIECE> > &candSol, int group){
	int i, expanse = 0;
	//Calculate the expase of the shapes in "group" to see if there is overflow
	for(i=0;i<candSol[group].size();i++)
	{
		if(i==0){ 
			//For the first shape add the projection on the left and the central bit
			if(candSol[group][i].LtoR) {
				expanse += (shape[candSol[group][i].num].p1 + shape[candSol[group][i].num].centralLen);
			}
			else {
				expanse += (shape[candSol[group][i].num].p2 + shape[candSol[group][i].num].centralLen);
			}
		}
		else{
			//Add the appropriate value according to the shape according to its allignment
			if(candSol[group][i-1].LtoR && candSol[group][i].LtoR){
				expanse += max(shape[candSol[group][i-1].num].p2, shape[candSol[group][i].num].p1) + shape[candSol[group][i].num].centralLen; 
			}
			else if(!candSol[group][i-1].LtoR && candSol[group][i].LtoR){
				expanse += max(shape[candSol[group][i-1].num].p1, shape[candSol[group][i].num].p1) + shape[candSol[group][i].num].centralLen;
			}
			else if(candSol[group][i-1].LtoR && !candSol[group][i].LtoR){
				expanse += max(shape[candSol[group][i-1].num].p2, shape[candSol[group][i].num].p2) + shape[candSol[group][i].num].centralLen;
			}
			else {//(!candSol[group][i-1].LtoR && !candSol[group][i].LtoR)
				expanse += max(shape[candSol[group][i-1].num].p1, shape[candSol[group][i].num].p2) + shape[candSol[group][i].num].centralLen;
			}
		}
	}
	//Dont forget to add on the final rightmost triangle
	if(candSol[group].back().LtoR){
		expanse += shape[candSol[group].back().num].p2;
	}
	else { 
		expanse += shape[candSol[group].back().num].p1;
	}
	return expanse;
}

//-------------------------------------------------------------------------------------
inline
void sortGroupsDescendingArea(vector<vector<PIECE> > &S, vector<long> &A){
	//Sorts the stocks by area used (for display purposes)
	int i,j;
	long tempD;
	for(i = S.size()-1; i>=0; i--){
		for(j=1; j<=i; j++){
			if(A[j-1] < A[j]){
				//the vectors are swapped
				S[j-1].swap(S[j]);
				//And so are the stocks areas
				tempD = A[j-1];	A[j-1] = A[j]; A[j] = tempD;
			}
		}
	}
}

//-------------------------------------------------------------------------------------
void prettyPrintByType(vector<vector<int> > candSol, vector<long> stockUsage){
	//This prints a copy of the solution to the screen.
	int i, count = 0, group;
	cout<<"\n";
	for(group=0; group<candSol.size(); group++){
		cout<<"Bin: "<<setw(4)<<group<<setw(6)<<setprecision(3)<<stockUsage[group]/double(STOCK_AREA)<<"\t( ";
		if(candSol[group].empty()) cout<<"-)";
		else {
			i=0;
			while(i<candSol[group].size()){
				cout << candSol[group][i]<<" ";
				i++;
			}
			cout<<")";
			count = count + candSol[group].size();
		}
		cout<<endl;
	}
	cout<<"Number of Items = "<<count<<endl;
}
void prettyPrintByType(vector<vector<PIECE> > candSol, vector<long> stockUsage){
	//This prints a copy of the solution to the screen.
	int i, count = 0, group;
	cout<<"\n";
	for(group=0; group<candSol.size(); group++){
		cout<<"Bin: "<<setw(4)<<group<<setw(6)<<setprecision(3)<<stockUsage[group]/double(STOCK_AREA)<<"\t( ";
		if(candSol[group].empty()) cout<<"-)";
		else {
			i=0;
			while(i<candSol[group].size()){
				cout << shape[candSol[group][i].num].shapeNum <<" ";
				i++;
			}
			cout<<")";
			count = count + candSol[group].size();
		}
		cout<<endl;
	}
	cout<<"Number of Items = "<<count<<endl;
}

//-------------------------------------------------------------------------------------
void prettyPrintSolution(vector<vector<PIECE> > candSol, vector<long> stockUsage){
	//This prints a copy of the solution to the screen.
	int i, count = 0, group, shapeExpanse=0;
	cout<<"\n";
	long stockAreaCheck = 0;
	for(group=0; group<candSol.size(); group++){
		cout<<"Bin: "<<setw(4)<<group<<setw(6)<<setprecision(3)<<stockUsage[group]/double(STOCK_AREA)<<"\t(";
		if(candSol[group].empty()) cout<<"-)";
		else {
			i=0;
			while(i<candSol[group].size()-1){
				cout << "(" << candSol[group][i].num<< ",";
				if(candSol[group][i].LtoR)cout<<"-->), ";
				else cout<<"<--), ";
				i++;
			}
			cout << "(" << candSol[group][i].num<< ",";
			if(candSol[group][i].LtoR) cout<<"-->))";
			else cout<<"<--))";
			count = count + candSol[group].size();
		}
		cout<<endl;
	}
	cout<<"Number of Items = "<<count<<endl;
}

//-------------------------------------------------------------------------------------
void prettyPrintPopulation(vector<vector<vector<PIECE> > > &pop, vector<vector<long> > popStockUsage){
	int i;
	for(i=0;i<pop.size();i++){
		prettyPrintSolution(pop[i], popStockUsage[i]);
	}
}

//-------------------------------------------------------------------------------------
void prettyPrintSolutionProj(vector<vector<PIECE> > candSol, vector<long> stockUsage){
	//This prints the a copy of the solution to the screen w.r.t the projections.
	int i, count = 0, group, shapeExpanse=0;
	cout<<"\n";
	long stockAreaCheck = 0;
	for(group=0; group<candSol.size(); group++){
		cout<<"Bin: "<<setw(4)<<group<<setw(6)<<setprecision(3)<<stockUsage[group]/double(STOCK_AREA)<<"\t(";
		if(candSol[group].empty()) cout<<"-)\n";
		else {
			for(i=0; i<candSol[group].size()-1; i++){
				if(candSol[group][i].LtoR)
					cout<<"("<<shape[candSol[group][i].num].p1<<","<<shape[candSol[group][i].num].p2<<")";
				else 
					cout<<"("<<shape[candSol[group][i].num].p2<<","<<shape[candSol[group][i].num].p1<<")";
			}
			if(candSol[group][i].LtoR)
				cout<<"("<<shape[candSol[group][i].num].p1<<","<<shape[candSol[group][i].num].p2<<"))";
			else 
				cout<<"("<<shape[candSol[group][i].num].p2<<","<<shape[candSol[group][i].num].p1<<"))";
			count = count + candSol[group].size();
		}
		cout<<endl;
	}
	cout<<"Number of Items = "<<count<<endl;
}

//-------------------------------------------------------------------------------------
void prettyPrintOutputSolution(vector<vector<PIECE> > candSol, vector<long> stockUsage){
	//This prints the solution to a file in such a way that it can be read by the display program
	//It works with a copy of the solution and first sorts the bins accroding how full they are
	int i, g;

	//First a the groups are ordered w.r.t their fullness
	sortGroupsDescendingArea(candSol,stockUsage);

	//Make a vector which will hold all of the positions of the shapes (0,1,2,3)
	vector<vector<int> > position(candSol.size());
	for(i=0;i<position.size();i++){
		vector<int> temp;
		position[i] = temp;
	}

	for(g=0; g<candSol.size(); g++){
		//Assume the first shape has its base facing downwards
		if(candSol[g][0].LtoR) position[g].push_back(1);
		else position[g].push_back(2);

		//Now determine the positions of the remaining shapes in bin g
		for(i=1;i<candSol[g].size();i++){
			//Write out details of the remaining items in the bin
			if(	   ( position[g].back() == 1 && shape[candSol[g][i-1].num].parallelogram == 1 ) 
				|| ( position[g].back() == 2 && shape[candSol[g][i-1].num].parallelogram == 1 )
				|| ( position[g].back() == 0 && shape[candSol[g][i-1].num].parallelogram == 0 )
				|| ( position[g].back() == 2 && shape[candSol[g][i-1].num].parallelogram == 0 )  ){
				//The previous shapes allignment implies a downwards slope on the right hand side, 
				//so we determine the position of the next shape accordingly
				if(candSol[g][i].LtoR)
					position[g].push_back(0);
				else if (shape[candSol[g][i].num].parallelogram == 0)
					position[g].push_back(2);
				else 
					position[g].push_back(3);
			}
			else {
				//The previous shape has an upwards slope, so we determine the position
				//of the next shape accordingly
				if(candSol[g][i].LtoR && shape[candSol[g][i].num].parallelogram == 0)
					position[g].push_back(1);
				else if(!candSol[g][i].LtoR && shape[candSol[g][i].num].parallelogram == 0)
					position[g].push_back(3);
				else if(candSol[g][i].LtoR && shape[candSol[g][i].num].parallelogram == 1)
					position[g].push_back(1);
				else position[g].push_back(2);
				
			}
		}
	}

	//Now write out the results in the required format
	ofstream answer;
	answer.open("theAnswer.txt");
	answer<<numItems<<" \n";
	answer<<candSol.size()<<" \n";
	for(g=0; g<candSol.size(); g++){
		answer<<candSol[g].size()<<" ";
		for(i=0;i<candSol[g].size();i++){
			answer<<candSol[g][i].num+1<<" "<<position[g][i]<<" ";
		}
		answer<<"\n";
	}
}

//-------------------------------------------------------------------------------------
void checkSolution(vector<vector<PIECE> > &candSol, vector<long> &stockUsage)
{
	//This checks the validity of a solution:
	//It checks: there are no empty sets
	//         : each item occurs exactly once
	//         : no stock is overflowing
	//         : the solution and the weights array match
	int i, count = 0, group, shapeExpanse, diffItemCount = 0;
	long stockAreaCheck = 0;
	bool Ok = true;
	vector<int> present(numItems,0);
	for(group=0; group<candSol.size(); group++){
		stockAreaCheck = 0;
		if(candSol[group].empty()) {
			cout<<"Error: (Group "<<group<<") Empty Set\n";
			Ok = false;
		}
		else { 
			for(i=0;i<candSol[group].size();i++){
				stockAreaCheck += shape[candSol[group][i].num].area;
				present[candSol[group][i].num]++;
			}
			count = count + candSol[group].size();
			shapeExpanse = calcExpanse(candSol,group);
			if(stockUsage[group] != stockAreaCheck){
				cout<<"Error: (Group "<<group<<") Weights in stockUsage array and dynamic check don't match\n";
				Ok = false;
			}
			if(shapeExpanse > STOCK_LEN){
				cout<<"Error: (Group "<<group<<") Shape expanse = "<<shapeExpanse<<", which is larger than STOCK_LEN\n";
				Ok = false;
			}
		}
	}
	if(count != numItems){
		cout<<"Error: Not enough (or too many) items in the solution\n";
		Ok = false;
	}
	for(i=0;i<present.size();i++){
		if(present[i]==0) {
			cout<<"Error: item "<<i<<" not present \n";
			Ok = false;
		} else if (present[i]>1) {
			cout<<"Error: item "<<i<<" occurs "<<present[i]<<" times \n";
			Ok = false;
		}
	}
	if(Ok && verbosity>=1)	cout<<"Solution is fine!"<<endl;
}
