#include "packshapes.h"

extern long TotalNumSubProbs;
extern long BoundNoSolCnt;
extern long SuccGreedyCnt;
extern long SuccLSCnt;
extern long SuccEulerCnt;
extern long FitNoSolCnt;

extern long STOCK_LEN;
extern long STOCK_WID;
extern long STOCK_AREA;

extern int subProbMethod;
extern int verbosity;
extern int algType;
extern int localSearchFlag;
extern int orderHeuristic;
extern int repairHeuristic;

extern int numItems;
extern int numTypes;
extern int tMin;
extern vector<SHAPE> shape;
extern vector<TYPE> types;

//Containers used with the Euler and greedy packing algorithms
vector<int> Freq;
vector<int> Size;
set<int> pSizes;
vector<int> SPrime;
vector<vector<PIECE> > subSeq;
vector<int> S;
vector<int> T;
vector<PIECE> solStock;
vector<PIECE> UStock;

//*********************************************************************
inline
int maxVal(int a, int b){
	if(a>=b)
		return a;
	else return(b);
}
inline 
int minVal(int a, int b){
	if(a<=b)
		return a;
	else return(b);
}
inline 
void swap(int &a, int &b){
	int x; x=a; a=b; b=x;
}
inline
int D(int a, int b){
	return abs(a-b);
}
inline
int RHS(vector<PIECE> &Seq, int pos){
	if(Seq[pos].LtoR) return (shape[Seq[pos].num].p2);
	else return (shape[Seq[pos].num].p1);
}
inline
int LHS(vector<PIECE> &Seq, int pos){
	if(Seq[pos].LtoR) return (shape[Seq[pos].num].p1);
	else return (shape[Seq[pos].num].p2);
}
inline
int RHS(vector<PIECE> &Seq){
	return (RHS(Seq,Seq.size()-1));
}
inline
int LHS(vector<PIECE> &Seq){
	return (LHS(Seq,0));
}

//*********************************************************************
//Returns false if there is a projection with no complement, OR if there are less
//than two square ends. This is used in the case that the shapes we are trying
//to pack have an area equal to the stock area.
inline
bool allProjectionsHaveComplement(vector<int> &SList){
	int i, j, numSquare=0, N = SList.size();
	bool complement;

	for(i=0; i<N; i++){
		
		if(shape[SList[i]].p1 == 0 || shape[SList[i]].p2 == 0) numSquare++;
		
		//Look for complement of shape[SList[i]].p1
		complement = false;
		for(j=i+1; j<(i+N); j++){
			if( (shape[SList[i]].p1 == shape[SList[j%N]].p1) || (shape[SList[i]].p1 == shape[SList[j%N]].p2) ){
				complement = true;
				break;
			}
		}
		if(!complement) return false;
		
		//Look for complement of shape[SList[i]].p2
		complement = false;
		for(j=i+1; j<(i+N); j++){
			if( (shape[SList[i]].p2 == shape[SList[j%N]].p1) || (shape[SList[i]].p2 == shape[SList[j%N]].p2) ){
				complement = true;
				break;
			}
		}
		if(!complement) return false;

	}

	// if we are here then as long as there are >2 flat edges, we still know that the items
	// *may* fit together
	if(numSquare >=2) return true;
	else return false;

}

//*********************************************************************
//Returns the size of the largest minimal gap between 2 shapes in S. Note that if largestMinGap == 0 then
//this implies that all shapes have a complement. Also if largestMinGap == 0 and we know that there
//is at least one square end then we know that all edges have a complement (with each other, or with
//the ends of the stock. However, we do the squareEnds check in the "allProjectionsHaveComplement" function
//so that this current (more expensive) function is not called so much
inline
int LargestMinGap(vector<int> &SList){
	int i, j, N = SList.size(), mingap, gap, largestMinGap = INT_MIN;

	for(i=0; i<N; i++){
		mingap = INT_MAX;
		for(j=i+1; j<(i+N); j++){
			//ID min gap for shape i's p1. Stop if we find a gap of 0
			gap = minVal( abs(shape[SList[i]].p1 - shape[SList[j%N]].p1) , abs(shape[SList[i]].p1 - shape[SList[j%N]].p2));
			//cout<<"gap "<<gap<<endl;
			if(gap < mingap) 
				mingap = gap;
			if(mingap == 0) 
				break;		
		}
		if(mingap > largestMinGap) largestMinGap = mingap;
		mingap = INT_MAX;
		for(j=i+1; j<(i+N); j++){
			//ID min gap for p2. Stop if we find a gap of 0
			gap = minVal( abs(shape[SList[i]].p2 - shape[SList[j%N]].p1) , abs(shape[SList[i]].p2 - shape[SList[j%N]].p2));
			//cout<<"gap "<<gap<<endl;
			if(gap < mingap) 
				mingap = gap;
			if(mingap == 0) 
				break;		
		}
		if(mingap > largestMinGap) largestMinGap = mingap;
	}

	if(largestMinGap<0){
		cout<<"Error with largestMinGap function\n";
		exit(1);
	}
	return largestMinGap;
}

//*********************************************************************
//This is called if we're trying to put just one or 2 items on a stock (quick and easy)
inline
void solveTrivial(vector<int> &SList, vector<PIECE> &finalStock, bool &inserted){
	PIECE x,y;
	if(SList.size()==1){	
		//We checked earlier that all shapes smaller than stocks. So just add this shape and end
		x.num=SList[0];
		x.LtoR=true;
		finalStock.push_back(x);
		inserted = true;
		return;
	}
	if(SList.size() == 2){
		int opt1, opt2, opt3, opt4, min;
		opt1 = shape[SList[0]].p1 + maxVal(shape[SList[0]].p2,shape[SList[1]].p1) + shape[SList[1]].p2;
		opt2 = shape[SList[0]].p1 + maxVal(shape[SList[0]].p2,shape[SList[1]].p2) + shape[SList[1]].p1;
		opt3 = shape[SList[0]].p2 + maxVal(shape[SList[0]].p1,shape[SList[1]].p1) + shape[SList[1]].p2;
		opt4 = shape[SList[0]].p2 + maxVal(shape[SList[0]].p1,shape[SList[1]].p2) + shape[SList[1]].p1;
		min = minVal(minVal(opt1,opt2),minVal(opt3,opt4));
		if(min + shape[SList[0]].centralLen + shape[SList[1]].centralLen > STOCK_LEN){
			inserted = false;
			return;
		}
		if(min == opt1){
			x.num = SList[0]; x.LtoR = true;
			y.num = SList[1]; y.LtoR = true;
		} else if(min == opt2){
			x.num = SList[0]; x.LtoR = true;
			y.num = SList[1]; y.LtoR = false;
		} else if(min == opt3){
			x.num = SList[0]; x.LtoR = false;
			y.num = SList[1]; y.LtoR = true;
		} else {
			x.num = SList[0]; x.LtoR = false;
			y.num = SList[1]; y.LtoR = false;
		}
		finalStock.push_back(x);
		finalStock.push_back(y);
		inserted = true;
		return;
	} 
}

//*********************************************************************
//Below are the procedures for subProbMethod=1 (which is approximate). First the shapes are fitted using the 
//greedy method. Next, if there is overflow, we try to improve things by flipping each shape. 
inline
void improveStockPacking(vector<PIECE> &stock, int &span){
	int currLWaste, currRWaste, newLWaste, newRWaste, newspan = 0, i;
	if(stock.size() <= 1){ cout<<"Error: stock must have > 1 item to run improveStock procedure\n"; exit(1); }

	//Go through each shape in turn and see if flipping it will reduce the wastage to its right and left
	//Flip first shape?
	currLWaste = LHS(stock,0);
	currRWaste = D(RHS(stock,0),LHS(stock,1));
	newLWaste  = RHS(stock,0);
	newRWaste  = D(LHS(stock,0),LHS(stock,1));
	if(newLWaste + newRWaste < currLWaste + currRWaste){
		//flip it
		stock[0].LtoR = !stock[0].LtoR;
	}
	newspan += (LHS(stock,0) + shape[stock[0].num].centralLen);
	//Flip the inner shapes?
	for(i=1; i<stock.size()-1; i++){
		currLWaste = D(RHS(stock,i-1),LHS(stock,i));
		currRWaste = D(RHS(stock,i),LHS(stock,i+1));
		newLWaste  = D(RHS(stock,i-1),RHS(stock,i));
		newRWaste  = D(LHS(stock,i),LHS(stock,i+1));	
		if(newLWaste + newRWaste < currLWaste + currRWaste){
			//flip it
			stock[i].LtoR = !stock[i].LtoR;
		}
		newspan += ( maxVal(RHS(stock,i-1),LHS(stock,i)) + shape[stock[i].num].centralLen );
	}
	//Flip last shape?
	currLWaste = D(RHS(stock,i-1),LHS(stock,i));
	currRWaste = RHS(stock,i);
	newLWaste  = D(RHS(stock,i-1),RHS(stock,i));
	newRWaste  = LHS(stock,i);
	if(newLWaste + newRWaste < currLWaste + currRWaste){
		//flip it
		stock[i].LtoR = !stock[i].LtoR;
	}
	newspan += ( maxVal(RHS(stock,i-1),LHS(stock,i)) + shape[stock[i].num].centralLen );

	//Dont forget to add the last triangle of wastage to span
	newspan += RHS(stock,i);
	
	if(newspan > span){ cout<<"Error: in improveStock procedure\n"; exit(1); }
	span = newspan;
}
inline
void greedy(vector<int> &SList, vector<PIECE> &finalStock, bool &inserted, bool allowOverflow){
	if(SList.size()<=0){ cout<<"Error: Set given to fit is empty...\n";exit(1);}
	
	int i, diff, s, mindiff, bestproj, minPos,  spare = STOCK_LEN, RHS = 0, span = 0;
	bool flipped, overflowing = false;
	PIECE thePiece;
		
	//Do some quick checks
	if(SList.size()<=2){
		solveTrivial(SList,finalStock,inserted);
		if(inserted) SuccGreedyCnt++;
		else FitNoSolCnt++;
		return;
	}
	
	SPrime = SList;
	//GREEDY ALGORITHM (NEAREST NEIGBOR)---------------
	while(!SPrime.empty()){
		//ID p1 or p2 in SPrime that is closest to RHS. Call this shape s
		mindiff = INT_MAX;
		for(i=0; i<SPrime.size(); i++){
			diff = abs(RHS - shape[SPrime[i]].p1);
			if(diff < mindiff){
				mindiff = diff; minPos = i; flipped = false; bestproj = shape[SPrime[i]].p1;
			}
			diff = abs(RHS - shape[SPrime[i]].p2);
			if(diff < mindiff){
				mindiff = diff; minPos = i; flipped = true; bestproj = shape[SPrime[i]].p2;
			}
		}
		//Now add s, keeping track of the span
		s = SPrime[minPos];
		span += maxVal(bestproj,RHS) + shape[s].centralLen; 

		//Update RHS
		if(!flipped) RHS = shape[s].p2;
		else RHS = shape[s].p1;
		
		//And add the shape to the stock
		thePiece.num=s;
		thePiece.LtoR=(!flipped);
		finalStock.push_back(thePiece);
		SPrime.erase(SPrime.begin() + minPos);	
	}
	//END OF MAIN LOOP------------------
	
	//Add last triangle onto Span
	if(finalStock.back().LtoR) 
		span += shape[finalStock.back().num].p2;
	else span += shape[finalStock.back().num].p1;

	if(span <= int(STOCK_LEN)){
		//If we are here then all of the shapes have been fitted by greedy without overflow.
		SuccGreedyCnt++;
		inserted = true;
		return;
	}
	
	//If we are here there is overflow so we attempt to improve the infeasible packing. 
	improveStockPacking(finalStock,span);
	if(span <= int(STOCK_LEN)){
		//improveStockPacking has resulted in a feasible packing so we end
		SuccLSCnt++;
		inserted = true;
		return;
	}
	
	//If we are here then we are still not able to pack the pieces onto the stock, we thus give up
	FitNoSolCnt++;
	inserted = false;
	return;
}

//*********************************************************************
//Below are the functions/procedures used with the *exact* Euler Algorithm
inline
long calcWastage(vector<PIECE> &Seq){
	int i;
	long wastage = 0;
	for(i=0;i<Seq.size()-1;i++){
		wastage += D(RHS(Seq,i),LHS(Seq,i+1));
	}
	wastage += D(RHS(Seq,i),LHS(Seq,0));
	return wastage;
}
inline
bool isOdd(int x){
	if(x%2==1)return true;
	else return false;
}
inline
void invert(vector<PIECE> &Seq){ 
	//Invert sequence
	int p1=0, p2=Seq.size()-1;
	while(true){
		if(p1 == p2) {
			//Need to flip center of inversion sequence
			Seq[p1].LtoR = !Seq[p1].LtoR;
			return;
		}
		if(p1 > p2) return;
		swap(Seq[p1],Seq[p2]);
		Seq[p1].LtoR = !Seq[p1].LtoR;
		Seq[p2].LtoR = !Seq[p2].LtoR;
		p1++;
		p2--;	
	}
}
inline
void mergeProjections(int i, bool iL, int j, bool jL){
	if(!iL && jL){
		//Merging right side of i to leftside of j
		subSeq[i].insert(subSeq[i].end(), subSeq[j].begin(), subSeq[j].end());
	} else if(!iL && !jL){
		//Merging right side of i to rightside of j
		invert(subSeq[j]);
		subSeq[i].insert(subSeq[i].end(), subSeq[j].begin(), subSeq[j].end());
	} else if(iL && jL){
		//Merging left side of i to leftside of j
		invert(subSeq[i]);
		subSeq[i].insert(subSeq[i].end(), subSeq[j].begin(), subSeq[j].end());
	} else {
		//Merging left side of i to right of j
		invert(subSeq[i]);
		invert(subSeq[j]);
		subSeq[i].insert(subSeq[i].end(), subSeq[j].begin(), subSeq[j].end());
	}
}
inline
int binSearch(vector<int> &A, int x){
	int L=0, U=A.size(), pos = (L+U)/2;
	while(A[pos] != x && L <= U){
		if(A[pos] > x) 
			U = pos-1;                               
		else 
			L = pos + 1;     
		pos = (L+U)/2;
	}
	if(L <= U)
		return pos;
	else
		return -1;
}
inline
void makeFreqTable(vector<int> &SList){
	//Make a set containing all of the different projection sizes that occur in the instance
	int i;
	std::set<int>::iterator it;
	for(i=0;i<SList.size();i++){
		pSizes.insert(shape[SList[i]].p1);
		pSizes.insert(shape[SList[i]].p2);
	}
	//Now use this information to build a table of frequencies
	Freq.resize(pSizes.size(),0);
	Size.resize(pSizes.size());
	i=0;
	for(it=pSizes.begin(); it!=pSizes.end(); it++) 
		Size[i++] = *it;
	for(i=0;i<SList.size();i++){
		Freq[binSearch(Size,shape[SList[i]].p1)]++;
		Freq[binSearch(Size,shape[SList[i]].p2)]++;
	}
}
inline
void Join(int x, int y, int fx, int fy){
	//Finds a projection of size x and another of size y and joins them if they are on different shapes.
	int i,j,opX,opY;
	bool iL, jL;
	for(i=0; i<subSeq.size();i++){
		if(LHS(subSeq[i]) == x){
			iL = true;
			opX = RHS(subSeq[i]);
			break;
		}
		if(RHS(subSeq[i]) == x){
			iL = false; 
			opX = LHS(subSeq[i]);
			break;
		}
	}
	for(j=0; j<subSeq.size();j++){
		if(LHS(subSeq[j]) == y){
			jL = true; 
			opY = RHS(subSeq[i]);
			break;
		}
		if(RHS(subSeq[j]) == y){
			jL = false; 
			opY = LHS(subSeq[i]);
			break;
		}
	}
	if(i>=subSeq.size() || j>=subSeq.size()){cout<<"Error here..\n"; exit(1);}
	if(i!=j){
		//Merge row j into row i and delete row j
		mergeProjections(i,iL,j,jL);
		subSeq[j].swap(subSeq.back());
		subSeq.pop_back();
	}
}
inline
void splice(vector<PIECE> &Seq1, vector<PIECE> &Seq2, int i, int j, bool invertNeeded){
	//First, rotate Seq2 so that the correct projections (the right-hand projection on jth item and the left-hand projection on the j+1th item) are on the outside
	rotate(Seq2.begin(), Seq2.begin()+((j+1)%Seq2.size()), Seq2.end());
	//Do an invert if needed
	if(invertNeeded) invert(Seq2);
	//Now insert Seq2 into the correct position in Seq1
	Seq1.insert(Seq1.begin()+i+1, Seq2.begin(), Seq2.end());
}
inline
void splice(vector<PIECE> &Seq1, vector<PIECE> &Seq2, int i, int j){
	//First, rotate Seq2 so that the correct projections (the right-hand projection on jth item and the left-hand projection on the j+1th item) are on the outside
	rotate(Seq2.begin(), Seq2.begin()+((j+1)%Seq2.size()), Seq2.end());
	//Now insert Seq2 into the correct position in Seq1
	Seq1.insert(Seq1.begin()+i+1, Seq2.begin(), Seq2.end());
}
inline
void matchZeroSequences(){
	int i,j,x,y,r1,r2,l1,l2;
	for(i=0;i<subSeq.size()-1;i++){
		for(j=0;j<subSeq[i].size();j++){
			r1=RHS(subSeq[i],j);
			l1=LHS(subSeq[i],(j+1)%subSeq[i].size());
			//Look for a pair of adjacent projections in a different subsequence that's equal to r1 and l1. Joining these brings no further wastage
start:
			for(x=i+1;x<subSeq.size();x++){
				for(y=0;y<subSeq[x].size();y++){
					r2=RHS(subSeq[x],y);
					l2=LHS(subSeq[x],(y+1)%subSeq[x].size());
					if(r1==r2 && l1==l2 && r1==l1){
						splice(subSeq[i],subSeq[x],j,y);
						subSeq[x].swap(subSeq.back());
						subSeq.pop_back();
						goto start;
					}
				}
			}
		}
	}
}
inline
void matchMinSequenceGeneralised(){
	int i,j,x,y,r1,l1,r2,l2,w1,w2,dist,mini,minj,minx,miny,min=INT_MAX;
	bool needInvert = false;
	//Identify the two closest pairs of adjacent projections in different sequences
	for(i=0;i<subSeq.size()-1;i++){
		for(j=0;j<subSeq[i].size();j++){
			r1=RHS(subSeq[i],j);
			l1=LHS(subSeq[i],(j+1)%subSeq[i].size());
			w1=D(r1,l1);
			for(x=i+1;x<subSeq.size();x++){
				for(y=0;y<subSeq[x].size();y++){
					r2=RHS(subSeq[x],y);
					l2=LHS(subSeq[x],(y+1)%subSeq[x].size());
					w2=D(r2,l2);
					dist = minVal(D(r1,r2)+D(l1,l2)-(w1+w2), D(r1,l2)+D(r2,l1)-(w1+w2));
					if(dist<min){
						min = dist;
						mini = i;
						minj = j;
						minx = x;
						miny = y;
						if(min == D(r1,r2)+D(l1,l2)-(w1+w2)) needInvert = true;
					}
				}
			}
		}
	}
	//The 2 closest projections are at subSeq[i][j] and subSeq[x][y]. So splice them together. Note that min can be less than 0
	splice(subSeq[mini],subSeq[minx],minj,miny,needInvert);
	subSeq[minx].swap(subSeq.back());
	subSeq.pop_back();
}
inline
void mergeMatchingProjectionsSequential(){
	//Now merge any shapes that have equal projections. (Do this in a sequential way). This is like IDing a buncn of Eulerian sub-circuits
	int i, j, selectedJ, op1, op2;
	bool foundMatch, iL, selectedJL;
	for(i=0; i<subSeq.size()-1;i++){
		while(true){
			//Look to see if there are any shapes that have a projection matching either of shape i's. Break ties by taking the onew with the biggest difference in the opposites
			foundMatch = false;
			for(j=i+1;j<subSeq.size();j++){
				if(LHS(subSeq[i]) == LHS(subSeq[j])){
					//Found matching projection between LHS (on row i) and LHS (on row j)
					iL = true; op1 = RHS(subSeq[i]); selectedJL = true; selectedJ = j; op2 = RHS(subSeq[j]); foundMatch = true; break;
				} 
				if (LHS(subSeq[i]) == RHS(subSeq[j])){
					//Found matching projection between LHS (on row i) and RHS (on row j)
					iL = true; op1 = RHS(subSeq[i]); selectedJL = false; selectedJ = j; op2 = LHS(subSeq[j]);foundMatch = true; break;
				}
				if (RHS(subSeq[i]) == LHS(subSeq[j])){
					//Found matching projection between RHS (on row i) and LHS (on row j)
					iL = false; op1 = LHS(subSeq[i]); selectedJL = true; selectedJ = j; op2 = RHS(subSeq[j]); foundMatch = true; break;
				}
				if (RHS(subSeq[i]) == RHS(subSeq[j])){
					//Found matching projection between RHS (on row i) and RHS (on row j) so merge and delete row j
					iL = false; op1 = LHS(subSeq[i]); selectedJL = false; selectedJ = j; op2 = LHS(subSeq[j]); foundMatch = true; break;
				}
			}
			if(foundMatch){
				//Merge row selectedJ into row i and delete row selectedJ
				mergeProjections(i,iL,selectedJ,selectedJL);
				subSeq[selectedJ].swap(subSeq.back());
				subSeq.pop_back();
			} 
			else break; //No match found for shape i
		}
	}
}
inline
void Euler(vector<int> &SList, vector<PIECE> &finalStock, bool &inserted, long spareArea){
	//Do some simple checks to start
	if(SList.size()<=0){
		cout<<"Error: Set given to fit is empty...\n";
		exit(1);
	} 
	if(SList.size()<=2){
		solveTrivial(SList,finalStock,inserted);
		if(inserted) SuccEulerCnt++;
		else FitNoSolCnt++;
		return;
	}
	
	int i,j,y;
	Freq.clear();
	Size.clear();
	pSizes.clear();

	//First, add the (0,0) dummy shape to SList
	SList.push_back(numItems);

	//Make a frequency table counting the number of projections of each size
	makeFreqTable(SList);

	//Now make initial array of subsequences (each subsequence is one item to start).
	subSeq.clear();
	subSeq.resize(SList.size());
	PIECE x;
	for(i=0;i<SList.size();i++){
		x.num = SList[i];
		x.LtoR = true;
		subSeq[i].push_back(x);
	}

	//Now join up pairs of one occurence of each odd projection with a dummy edge. This makes the freq/degree of all shapes even.
	for(i=0;i<Freq.size()-1;i++){
		if(isOdd(Freq[i])){
			for(j=i+1;j<Freq.size();j++){
				if(isOdd(Freq[j])){
					//Size[i] and Size[j] have an odd number of projections
					Join(Size[i],Size[j],Freq[i],Freq[j]);
					break;
				}
			}
			i=j;
		}
	}

	//At this point, the frequency of all exterior projections is even. So we run the Euler tour algorithm. This does not incur any wastage
	if(subSeq.size() > 1) mergeMatchingProjectionsSequential();
	//...splicing together overlapping subcircuits (as with Hierholzer's algorithm)
	if(subSeq.size() > 1) matchZeroSequences();

	//We now have one or more disjoint Eulerian circuits. These also need to be spliced together in a minimal-cost way.
	//Do this by repeatedly splicing together the Eulerian sub-circuits that incur minimal wastage at each step (this incurs further wastage)
	while(subSeq.size() > 1){
		matchMinSequenceGeneralised();
	}

	//Now calculate the amount of inter-item wastage on the resultant stock. If this is more than the spare area then we cannot pack these items.
	if(calcWastage(subSeq[0]) > spareArea){
		FitNoSolCnt++;
		inserted = false;
		return;
	}

	//If we are here we have the optimal sequence in subSeq[0], and this will fit on to the stock
	//Rearrange so (0,0) is at the back, then delete it, then copy the rordered items into final stock and end
	SuccEulerCnt++;
	inserted = true;
	finalStock.clear();
	for(y=0;y<subSeq[0].size();y++) if(subSeq[0][y].num == numItems) break;
	if(y <= subSeq[0].size()){
		for(i=y+1;i<subSeq[0].size();i++)	finalStock.push_back(subSeq[0][i]);
		for(i=0;i<y;i++)					finalStock.push_back(subSeq[0][i]); //NB: We ignore subSeq[0][x] as this is the (0,0) item
	}
}

//*********************************************************************
bool packingPossible(int item, vector<vector<PIECE> > &candSol, int g, vector<long> &stockUsage, vector<PIECE> &finalStock)
{
	//If the packing is possible then the item is added and we return true, else nothing is changed and we return false 
	int i;

	//Calculate amount of spare area on the stock 
	long spareArea = STOCK_AREA - (stockUsage[g] + shape[item].area);
	if(spareArea < 0){
		//Not enough overall area so we immediately know it cannot be packed
		return false;
	}
	
	TotalNumSubProbs++;
	//We have confirmed that there is enough space IN TOTAL for "item", to be accomodated and we now need to check whether it can actually fit there. 
	//We first make a set S containing all the shapes we want to pack
	S.clear();
	for(i=0;i<candSol[g].size();i++) S.push_back(candSol[g][i].num);
	S.push_back(item);

	//Now we do some quick checks to see if the shapes will fit (these are just bound checks). We exit immediately if these are not satisfied
	if(spareArea == 0){
		if(!allProjectionsHaveComplement(S)){
			BoundNoSolCnt++;
			return false;
		}
	} 
	if(spareArea < long((LargestMinGap(S)*STOCK_WID)/2)){
		BoundNoSolCnt++;
		return false;
	}

	//We have decided that the shapes in S may well fit together. So make a placeholder stock that will replace the existing one if S can be feasibly packed
	finalStock.clear();
	bool inserted = false;

	if(subProbMethod==1){
		//We run the linear, greedy shape fitting algorithm + flip operator (this method is approximate)
		greedy(S,finalStock,inserted,true);	
	} else {
		//Otherwise the exact method is applied (this method is exact)
		Euler(S,finalStock,inserted,spareArea);
	}

	//If both new groups are fine, we can update the solution. Otherwise we end.
	if(inserted){
		candSol[g].swap(finalStock);
		stockUsage[g] += shape[item].area;
		return true;
	}
	else { 
		//The item cannot go in this group and nothing is changed
		return false;
	}
}

//*********************************************************************
bool doPairPairSwap(vector<vector<PIECE> > &partialSol, int g, int j, int ka, vector<vector<PIECE> > &U, int h, int m, int n,	long pairSize, long uSize, vector<long> &stockUsage)
{
	int i;
	
	//Now do a check to see whether we are swapping the only two items in a the partialSol bin with the only two
	//items in the unplaced bin. Obviously this swap will result in a feasible groups so no further checks are needed
	if(partialSol[g].size() == 2 && U[h].size() == 2){
		partialSol[g].swap(U[h]);
		stockUsage[g]=uSize;
		return true;
	}
	
	long spareArea = STOCK_AREA - ((stockUsage[g]-pairSize) + uSize);
	//Now we want to check that swapping the pairs would allow both stocks to remain feasible
	//We first make two sets S and T containg all the shapes we want to pack
	S.clear();
	for(i=0;i<partialSol[g].size();i++){
		if(i != j && i != ka) S.push_back(partialSol[g][i].num);
	}
	S.push_back(U[h][m].num);
	S.push_back(U[h][n].num);

	T.clear();
	for(i=0;i<U[h].size();i++){
		if(i != m && i != n) T.push_back(U[h][i].num);
	}
	T.push_back(partialSol[g][j].num);
	T.push_back(partialSol[g][ka].num);

	//Calculate spare area on the stock define by T too (by the nature of this operator, UspareArea should always be positive)
	long USpareArea, tot=0;
	for(i=0;i<T.size();i++)tot+=shape[T[i]].area;
	USpareArea = STOCK_AREA - tot;

	//We now need to check that both S and T will make feasible packings
	//and make 2 placeholder stocks that will replace the existing ones if they can
	solStock.clear();
	UStock.clear();
	bool partialOK = false, UOK = false;
	
	//Here we check whether the shapes in S will need to fit exactly. If so we can do a quick check for projection 
	//complements. We only do this check for S since it it the only group whose size is increasing 
	//(and will thus fit this criteria). We can also check whether the largest minimal gap is bigger than the spare.
	TotalNumSubProbs++;
	if(spareArea == 0){
		if(!allProjectionsHaveComplement(S)){
			BoundNoSolCnt++;
			return false;
		}
	} 
	if(spareArea < long((LargestMinGap(S)*STOCK_WID)/2)){
		//LargestMinGap bigger than total spare area so we know te shapes will not fit
		BoundNoSolCnt++;
		return false;
	}

	if(subProbMethod==1){
		greedy(S,solStock,partialOK,true);
	} else {
		Euler(S,solStock,partialOK,spareArea);
	}

	//Now check T as well
	if(partialOK){
		//That swap was fine so check the other one too
		TotalNumSubProbs++;
		if(USpareArea < long((LargestMinGap(T)*STOCK_WID)/2)){
			BoundNoSolCnt++;
			return false;
		}
		if(subProbMethod==1){		
			greedy(T,UStock,UOK,true);
		} else {
			Euler(T,UStock,UOK,USpareArea);	
		}
	} 

	//If both new groups are fine, we can update the solution. Otherwise we end.
	if(partialOK && UOK){
		partialSol[g].swap(solStock);
		U[h].swap(UStock);
		stockUsage[g] = (stockUsage[g]-pairSize)+uSize;
		return true;
	}
	else return false;
}
	
//*********************************************************************
bool doPairSinSwap(vector<vector<PIECE> > &partialSol, int g, int j, int ka, vector<vector<PIECE> > &U, int h, int m, long pairSize, long uSize, vector<long> &stockUsage)
{
	int i;
	//Returns true if the swap can be made, and also performs all the relevant changes; Otherwise returns false and changes nothing
	
	//Do a check to see whether we are swapping the only two items in the partialSol bin with the only 
	//item in the unplaced bin. Obviously this will result in a feasible groups so no further checks are needed
	if(partialSol[g].size() == 2 && U[h].size() == 1){
		partialSol[g].swap(U[h]);
		stockUsage[g]=uSize;
		return true;
	}
	
	long spareArea = STOCK_AREA - ((stockUsage[g]-pairSize) + uSize);
	//Form S and T
	S.clear();
	T.clear();
	for(i=0;i<partialSol[g].size();i++){
		if(i != j && i != ka) S.push_back(partialSol[g][i].num);
	}
	S.push_back(U[h][m].num);
	
	for(i=0;i<U[h].size();i++){
		if(i != m) T.push_back(U[h][i].num);
	}
	T.push_back(partialSol[g][j].num);
	T.push_back(partialSol[g][ka].num);

	//Calculate spare area on the stock defined by T
	long USpareArea, tot=0;
	for(i=0;i<T.size();i++)tot+=shape[T[i]].area;
	USpareArea = STOCK_AREA - tot;

	//We now need to check that both S and T will make feasible packings
	//and make 2 placeholder stocks that will replace the existing ones if they can
	solStock.clear();
	UStock.clear();
	bool partialOK = false, UOK = false;

	//Here we check whether the shapes in S will need to fit exactly. If so we can do a quick check for projection 
	//complements. We only do this check for S since it it the only group whose size is increasing 
	//(and will thus fit this criteria). We can also check whether the largest minimal gap is bigger than the spare.
	TotalNumSubProbs++;
	if(spareArea == 0){
		if(!allProjectionsHaveComplement(S)){
			BoundNoSolCnt++;
			return false;
		}
	}
	if(spareArea < long((LargestMinGap(S)*STOCK_WID)/2)){
		//LargestMinGap bigger than total spare area so we know te shapes will not fit
		BoundNoSolCnt++;
		return false;
	}
	
	if(subProbMethod==1){
		greedy(S,solStock,partialOK,true);
	} else {
		Euler(S,solStock,partialOK,spareArea);
	}

	//Now check T as well
	if(partialOK){
		//That swap was fine so check the other one too
		TotalNumSubProbs++;
		if(USpareArea < long((LargestMinGap(T)*STOCK_WID)/2)) {
			BoundNoSolCnt++;
			return false;
		}
		if(subProbMethod==1){
			greedy(T,UStock,UOK,true);
		} else {
			Euler(T,UStock,UOK,USpareArea);	
		}
	}

	//If both new groups are fine, we can update the solution. Otherwise we end.
	if(partialOK && UOK){
		partialSol[g].swap(solStock);
		U[h].swap(UStock);
		stockUsage[g] = (stockUsage[g]-pairSize)+uSize;
		return true;
	}
	else return false;

}
	
//*********************************************************************
bool doSinSinSwap(vector<vector<PIECE> > &partialSol, int g, int j, vector<vector<PIECE> > &U, int h, int m, long sinSize, long uSize, vector<long> &stockUsage)
{
	int i;
	//Returns true if the swap can be made, and also performs all the relevant changes;
	//Otherwise returns false and changes nothing
	
	//Now do a check to see whether we are swapping the only item in the partialSol bin with the only 
	//items in the unplaced bin. Obviously this will result in a feasible groups so no further checks are needed
	if(partialSol[g].size() == 1 && U[h].size() == 1){
		partialSol[g].swap(U[h]);
		stockUsage[g]=uSize;
		return true;
	}
	

	long spareArea = STOCK_AREA - ((stockUsage[g]-sinSize) + uSize);
	S.clear();
	T.clear();
	for(i=0;i<partialSol[g].size();i++){
		if(i != j) S.push_back(partialSol[g][i].num);
	}
	S.push_back(U[h][m].num);

	for(i=0;i<U[h].size();i++){
		if(i != m) T.push_back(U[h][i].num);
	}
	T.push_back(partialSol[g][j].num);

	//Calculate spare area on the stock define by T too (by the nature of this operator, UspareArea should always be positive)
	long USpareArea, tot=0;
	for(i=0;i<T.size();i++)tot+=shape[T[i]].area;
	USpareArea = STOCK_AREA - tot;

	//We now need to check that both S and T will make feasible packings
	//and make 2 placeholder stocks that will replace the existing ones if they can
	solStock.clear();
	UStock.clear();
	bool partialOK = false, UOK = false;

	//Here we check whether the shapes in S will need to fit exactly. If so we can do a quick check for projection 
	//complements. We only do this check for S since it it the only group whose size is increasing 
	//(and will thus fit this criteria). We can also check whether the largest minimal gap is bigger than the spare.
	TotalNumSubProbs++;
	if(spareArea==0){
		if(!allProjectionsHaveComplement(S)){
			BoundNoSolCnt++;
			return false;
		}
	} 
	if(spareArea < long((LargestMinGap(S)*STOCK_WID)/2)){
		//LargestMinGap bigger than total spare area so we know te shapes will not fit
		BoundNoSolCnt++;
		return false;
	}
	
	if(subProbMethod==1){
		greedy(S,solStock,partialOK,true);
	} else {
		Euler(S,solStock,partialOK,spareArea);
	}

	if(partialOK){
		//That swap was fine so check the other one too
		TotalNumSubProbs++;
		if(USpareArea < long((LargestMinGap(T)*STOCK_WID)/2)) {
			BoundNoSolCnt++;
			return false;
		}
		if(subProbMethod==1){
			greedy(T,UStock,UOK,true);
		} else {
			Euler(T,UStock,UOK,USpareArea);
		}
	}

	//If both new groups are fine, we can update the solution. Otherwise we end.
	if(partialOK && UOK){
		partialSol[g].swap(solStock);
		U[h].swap(UStock);
		stockUsage[g] = (stockUsage[g]-sinSize)+uSize;
		return true;
	}
	else return false;
}
